# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for supervised training: core classes and flows."""

import collections
import os
import time

from absl.testing import absltest
from jax import test_util  # pylint: disable=unused-import
from jax.config import config
import numpy as np

from trax import data
from trax import fastmath
from trax import layers as tl
from trax import optimizers
from trax import shapes
from trax import test_utils
from trax.layers import base
from trax.models import transformer
from trax.supervised import callbacks
from trax.supervised import training


class TrainingTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')

  def test_loop_no_eval_task(self):
    """Runs a training loop with no eval task(s)."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    training_session = training.Loop(model, [task])
    # Loop should initialize and run successfully, even with no eval task.
    training_session.run(n_steps=5)


  def test_loop_checkpoint_low_metric(self):
    """Runs a training loop that saves checkpoints for low metric values."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(_very_simple_data(),
                              tl.L2Loss(),
                              optimizers.SGD(.01))
    eval_metric = tl.L2Loss()
    eval_task = training.EvalTask(_very_simple_data(),
                                  [eval_metric],
                                  metric_names=['l2_loss'])
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model,
                         [task],
                         eval_tasks=[eval_task],
                         output_dir=tmp_dir,
                         eval_at=lambda step_n: step_n % 2 == 0,
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         checkpoint_low_metric='l2_loss')
    call_counter = collections.Counter()
    loop.save_checkpoint = lambda name: call_counter.update([name])
    loop.run(n_steps=10)

    # Eval metric steadily descends, so low checkpoint triggered all 5 times.
    # High checkpoint not defined, so never triggered.
    self.assertEqual(call_counter['model'], 5)
    self.assertEqual(call_counter['lowest_l2_loss'], 5)
    self.assertEqual(call_counter['highest_l2_loss'], 0)

  def test_loop_checkpoint_high_metric(self):
    """Runs a training loop that saves checkpoints for high metric values."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(_very_simple_data(),
                              tl.L2Loss(),
                              optimizers.SGD(.01))
    eval_metric = tl.L2Loss()
    eval_task = training.EvalTask(_very_simple_data(),
                                  [eval_metric],
                                  metric_names=['l2_loss'])
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model,
                         [task],
                         eval_tasks=[eval_task],
                         output_dir=tmp_dir,
                         eval_at=lambda step_n: step_n % 2 == 0,
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         checkpoint_high_metric='l2_loss')
    call_counter = collections.Counter()
    loop.save_checkpoint = lambda name: call_counter.update([name])
    loop.run(n_steps=10)

    # Eval metric steadily descends, so high checkpoint triggered only once.
    # Low checkpoint not defined, so never triggered.
    self.assertEqual(call_counter['model'], 5)
    self.assertEqual(call_counter['lowest_l2_loss'], 0)
    self.assertEqual(call_counter['highest_l2_loss'], 1)

  def test_train_dense_layer(self):
    """Trains a very simple network on a very simple task."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    training_session = training.Loop(model, [task], eval_tasks=[eval_task],
                                     eval_at=lambda step_n: step_n % 2 == 0)
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=15)
    self.assertEqual(15, training_session.step)
    training_session.run(n_steps=5)
    self.assertEqual(20, training_session.step)

  def test_loop_with_initialized_model(self):
    """Check that loop does not re-initialize an already initialized model."""
    model = tl.Serial(tl.Dense(1))
    example_data = next(_very_simple_data())
    model.init(example_data)
    w = model.weights[0][0]
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    loop = training.Loop(model, [task], eval_tasks=[eval_task],
                         eval_at=lambda step_n: step_n % 2 == 0)
    self.assertEqual(0, loop.step)
    self.assertEqual(loop.model.weights[0][0], w)

  def test_train_save_restore_dense(self):
    """Saves and restores a checkpoint to check for equivalence."""
    train_data = data.Serial(lambda _: _very_simple_data(),
                             data.CountAndSkip('simple_data'))
    task = training.TrainTask(
        train_data(), tl.L2Loss(), optimizers.Adam(.0001))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    tmp_dir = self.create_tempdir().full_path

    def _make_model_and_session():
      m = tl.Serial(tl.Dense(1))
      ts = training.Loop(m, [task], eval_tasks=[eval_task],
                         eval_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
      return m, ts

    model, training_session = _make_model_and_session()
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=1)
    training_session.save_checkpoint('model')
    self.assertEqual(data.inputs.data_counters['simple_data'], 2)
    data.inputs.data_counters['simple_data'] = 0  # reset manually
    self.assertEqual(data.inputs.data_counters['simple_data'], 0)  # check
    model2, training_session2 = _make_model_and_session()
    self.assertEqual(data.inputs.data_counters['simple_data'], 2)  # restored

    x = np.ones((8, 1))
    y1 = model(x, rng=fastmath.random.get_prng(0))
    y2 = model2(x, rng=fastmath.random.get_prng(0))
    self.assertEqual(str(y1), str(y2))

    training_session2.run(n_steps=1)
    y1 = model(x, rng=fastmath.random.get_prng(0))
    y2 = model2(x, rng=fastmath.random.get_prng(0))
    self.assertNotEqual(str(y1), str(y2))

    slots1 = training_session._trainer_per_task[0].slots
    slots2 = training_session2._trainer_per_task[0].slots
    np.testing.assert_array_equal(slots1, slots2)

  def test_train_save_restore_sharded(self):
    """Saves and restores a sharded checkpoint to check for equivalence."""
    if fastmath.local_device_count() < 2:
      return  # multi-accelerator only
    base.N_WEIGHTS_SHARDS = fastmath.local_device_count()
    train_data = data.Serial(lambda _: _very_simple_data(2, 2),
                             data.CountAndSkip('simple_data'))
    task = training.TrainTask(
        train_data(), tl.L2Loss(), optimizers.Adam(.0001))
    eval_task = training.EvalTask(
        _very_simple_data(2, 2),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    tmp_dir = self.create_tempdir().full_path

    def _make_model_and_session():
      m = tl.Serial(tl.Dense(2))
      ts = training.Loop(m, [task], eval_tasks=[eval_task],
                         eval_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
      return m, ts

    _, training_session = _make_model_and_session()
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=1)
    training_session.save_checkpoint('model')
    _, training_session2 = _make_model_and_session()
    training_session2.run(n_steps=1)
    base.N_WEIGHTS_SHARDS = 1

  def test_train_save_restore_transformer(self):
    """Saves and restores a checkpoint to check for equivalence."""
    vocab_size = 8
    task = training.TrainTask(
        _very_simple_transformer_data(), tl.L2Loss(), optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_transformer_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    tmp_dir = self.create_tempdir().full_path

    def _make_model_and_session():
      m = transformer.TransformerLM(
          vocab_size, d_model=4, d_ff=4, n_layers=1, n_heads=2, dropout=0.)
      ts = training.Loop(m, [task], eval_tasks=[eval_task],
                         eval_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
      return m, ts

    model, training_session = _make_model_and_session()
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=1)
    training_session.save_checkpoint('model')
    model2, training_session2 = _make_model_and_session()

    x = np.ones((2, 2)).astype(np.int32)
    y1 = model(x, rng=fastmath.random.get_prng(0))
    y2 = model2(x, rng=fastmath.random.get_prng(0))
    self.assertEqual(str(y1), str(y2))

    training_session2.run(n_steps=1)
    y1 = model(x, rng=fastmath.random.get_prng(0))
    y2 = model2(x, rng=fastmath.random.get_prng(0))
    self.assertNotEqual(str(y1), str(y2))

  def test_train_dense_layer_with_momentum(self):
    """Trains with an optimizer that has slots / requires initialization."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.Momentum(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['Momentum.L2Loss'])
    training_session = training.Loop(model, [task], eval_tasks=[eval_task],
                                     eval_at=lambda step_n: step_n % 2 == 0)
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=20)
    self.assertEqual(20, training_session.step)

  def test_train_dense_layer_evals(self):
    """Trains a very simple network on a very simple task, 2 epochs."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()])
    training_session = training.Loop(model, [task], eval_tasks=[eval_task],
                                     eval_at=lambda step_n: False)
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=10)
    self.assertEqual(10, training_session.step)
    training_session.run_evals()
    self.assertEqual(10, training_session.step)  # Unchanged

  def test_summaries_are_written(self):
    """Training writes down metrics when writing is turned on."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        metric_names=['SGD.L2Loss'])
    tmp_dir = self.create_tempdir().full_path
    training_session = training.Loop(model, [task], eval_tasks=[eval_task],
                                     eval_at=lambda step_n: step_n % 2 == 0,
                                     output_dir=tmp_dir)
    expected_train_metric_dir = os.path.join(tmp_dir, 'train')
    expected_eval_metric_dir = os.path.join(tmp_dir, 'eval')
    for directory in [expected_train_metric_dir, expected_eval_metric_dir]:
      self.assertFalse(
          os.path.isdir(directory), 'Failed for directory %s.' % directory)
    training_session.run(n_steps=15)
    time.sleep(1)  # wait for the files to be closed
    for directory in [expected_train_metric_dir, expected_eval_metric_dir]:
      self.assertTrue(
          os.path.isdir(directory), 'Failed for directory %s.' % directory)
      self.assertEqual(
          1, _count_files(directory), 'Failed for directory %s.' % directory)
    training_session.run(n_steps=5)
    time.sleep(1)  # wait for the files to be closed
    for directory in [expected_train_metric_dir, expected_eval_metric_dir]:
      self.assertEqual(
          2, _count_files(directory), 'Failed for directory %s.' % directory)

  def test_restores_step(self):
    """Training restores step from directory where it saved it."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
    loop.run(4)
    loop2 = training.Loop(model, [task], output_dir=tmp_dir)
    self.assertEqual(4, loop2.step)

  def test_restores_memory_efficient_from_standard(self):
    """Training restores step from directory where it saved it."""
    model = tl.Serial(tl.Dense(4), tl.Dense(1))
    task_std = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.Adam(.0001))
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model, [task_std],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
    loop.run(4)
    task_memeff = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.Adam)
    loop2 = training.Loop(model, [task_memeff], output_dir=tmp_dir,
                          use_memory_efficient_trainer=True)
    loop2.run(2)
    self.assertEqual(6, loop2.step)

  def test_restores_from_smaller_model(self):
    """Training restores from a checkpoint created with smaller model."""
    model1 = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.Adam(.01))
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model1, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
    loop.run(2)
    model2 = tl.Serial(tl.Dense(1), tl.Dense(1))
    loop2 = training.Loop(model2, [task], output_dir=tmp_dir)
    self.assertEqual(2, loop2.step)

  def test_restore_fails_different_model(self):
    """Training restores from a checkpoint created with smaller model."""
    model1 = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01))
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model1, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
    loop.run(2)
    model2 = tl.Serial(tl.Dense(2))
    with self.assertRaises(IndexError):
      training.Loop(model2, [task], output_dir=tmp_dir)

  def test_restores_step_bfloat16(self):
    """Training restores step from directory where it saved it, w/ bfloat16."""
    model = tl.Serial(tl.Dense(1, use_bfloat16=True))
    # We'll also use Adafactor with bfloat16 to check restoring bfloat slots.
    opt = optimizers.Adafactor(.01, do_momentum=True, momentum_in_bfloat16=True)
    task = training.TrainTask(_very_simple_data(), tl.L2Loss(), opt)
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir)
    loop.run(4)
    loop2 = training.Loop(model, [task], output_dir=tmp_dir)
    self.assertEqual(4, loop2.step)
    loop2.run(2)  # check that continued training works
    self.assertEqual(6, loop2.step)

  def test_restores_step_sharded(self):
    """Training restores step from directory where it saved it, sharded."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD)
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir, use_memory_efficient_trainer=True)
    loop.run(4)
    loop2 = training.Loop(model, [task],
                          output_dir=tmp_dir, use_memory_efficient_trainer=True)
    self.assertEqual(4, loop2.step)

  def test_restores_step_sharded_bfloat16(self):
    """Training restores step from where it saved it, sharded and bfloat16."""
    model = tl.Serial(tl.Dense(1, use_bfloat16=True))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD)
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(model, [task],
                         checkpoint_at=lambda step_n: step_n % 2 == 0,
                         output_dir=tmp_dir, use_memory_efficient_trainer=True)
    loop.run(4)
    loop2 = training.Loop(model, [task],
                          output_dir=tmp_dir, use_memory_efficient_trainer=True)
    self.assertEqual(4, loop2.step)
    loop2.run(2)  # check that continued training works
    self.assertEqual(6, loop2.step)

  def test_restores_history(self):
    """Training restores history from directory where it saved it."""
    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(_very_simple_data(), tl.L2Loss(),
                              optimizers.SGD(.01))
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()])
    tmp_dir = self.create_tempdir().full_path
    loop = training.Loop(
        model, [task],
        eval_tasks=[eval_task],
        eval_at=lambda step_n: step_n % 2 == 0,
        checkpoint_at=lambda step_n: step_n % 2 == 0,
        output_dir=tmp_dir)
    loop.run(4)
    loop2 = training.Loop(model, [task], output_dir=tmp_dir)
    self.assertLen(loop2.history.modes, 2)
    self.assertLen(loop2.history.metrics_for_mode('train'), 6)
    self.assertLen(loop2.history.metrics_for_mode('eval'), 1)
    for mode, metric in [
        ('train', 'metrics/L2Loss'),
        ('train', 'training/learning_rate'),
        ('train', 'training/steps per second'),
        ('train', 'training/gradients_l2'),
        ('train', 'training/loss'),
        ('train', 'training/weights_l2'),
        ('eval', 'metrics/L2Loss'),
    ]:
      self.assertLen(loop2.history.get(mode, metric), 1)
      self.assertEqual(2, loop2.history.get(mode, metric)[0][0])

  def test_trains_on_two_tasks(self):
    """Trains a very simple network on two very simple tasks."""
    model = tl.Serial(tl.Dense(3), tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(),
        tl.L2Loss(),
        optimizers.SGD(.01)
    )
    eval_task = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
    )
    training_session = training.Loop(
        model,
        tasks=(task, task),
        eval_tasks=(eval_task, eval_task),
        which_task=lambda step_n: step_n % 2,
    )
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=15)
    self.assertEqual(15, training_session.step)
    training_session.run(n_steps=5)
    self.assertEqual(20, training_session.step)

  def test_train_one_task_eval_two_tasks(self):
    """Trains a very simple network on one task and evaluates on two tasks."""
    model = tl.Serial(tl.Dense(3), tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(),
        tl.L2Loss(),
        optimizers.SGD(.01)
    )
    export_prefix_1 = 'eval_1'
    eval_task_1 = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        export_prefix=export_prefix_1,
    )
    export_prefix_2 = 'eval_2'
    eval_task_2 = training.EvalTask(
        _very_simple_data(),  # deliberately re-using training data
        [tl.L2Loss()],
        export_prefix=export_prefix_2,
    )
    training_session = training.Loop(
        model,
        tasks=(task,),
        eval_tasks=(eval_task_1, eval_task_2),
    )
    self.assertEqual(0, training_session.step)
    training_session.run(n_steps=5)
    self.assertEqual(5, training_session.step)
    export_prefixes = [task.export_prefix
                       for task in training_session.eval_tasks]
    self.assertCountEqual([export_prefix_1, export_prefix_2],
                          export_prefixes)

  def test_can_predict_with_trained_model(self):
    model = tl.Serial(tl.Dense(3), tl.Branch(tl.Dense(1), tl.Dense(2)))
    train_tasks, eval_tasks = [], []
    for output_dim in [1, 2]:
      # The head we select from the model: 0 for output_dim 1 and 1 for 2.
      head_index = output_dim - 1
      train_tasks.append(training.TrainTask(
          _very_simple_data(output_dim),
          tl.Serial(tl.Select([head_index], n_in=2), tl.L2Loss()),
          optimizers.SGD(.01)
      ))
      eval_tasks.append(training.EvalTask(
          _very_simple_data(output_dim),  # deliberately re-use training data
          [tl.Serial(tl.Select([head_index], n_in=2), tl.L2Loss())]
      ))
    tmp_dir = self.create_tempdir().full_path
    training_session = training.Loop(
        model,
        tasks=train_tasks,
        eval_tasks=eval_tasks,
        checkpoint_at=lambda step_n: step_n == 1,
        output_dir=tmp_dir,
        which_task=lambda step_n: step_n % 2,
    )
    training_session.run(n_steps=2)

    trained_model = training_session.eval_model
    inp = next(_very_simple_data())[0]
    out = trained_model(inp)
    self.assertEqual(
        shapes.signature(out),
        (shapes.ShapeDtype((8, 1)), shapes.ShapeDtype((8, 2))),
    )

  def test_train_memory_efficient(self):
    """Trains a large network in a memory-efficient way."""
    # This test requires > 16GB RAM, only run on TPUs. It does pass on GPU
    # and CPU when you run it locally, but it's too big for unit-testing.
    ram_limited = True  # Set to False to run this test locally.
    if fastmath.global_device_count() == 1 and ram_limited:
      return

    # Create the model.
    n_layers = 16  # 16 layers each 16K x 16K = 256M weights ~= 1GB, 16GB ram
    model = tl.Serial(
        tl.Embedding(9, 16*1024),
        tl.Dup(),
        [[tl.ReversibleHalfResidual(tl.Dense(16*1024)), tl.ReversibleSwap()]
         for _ in range(n_layers)],
        tl.Concatenate(),
        tl.Dense(9),
    )

    # Create inputs.
    inputs_batch = np.arange(8).reshape((2, 4))
    targets_batch = inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    def _data_gen():
      while True:
        yield labeled_batch

    # Run training.
    loss_layer = tl.WeightedCategoryCrossEntropy()
    task = training.TrainTask(_data_gen(), loss_layer, optimizers.Adafactor)
    eval_task = training.EvalTask(_data_gen(),
                                  [tl.WeightedCategoryCrossEntropy()])
    loop = training.Loop(model, [task], eval_tasks=[eval_task],
                         eval_at=lambda step_n: step_n == 2,
                         use_memory_efficient_trainer=True)
    self.assertEqual(0, loop.step)
    loop.run(n_steps=2)
    self.assertEqual(2, loop.step)

  def test_initializes_step_callbacks_with_loop_instance(self):
    """Runs a training loop, asserting that callbacks are initialized."""

    class ActualLoop:
      # Wrapper object to make the Loop reference mutable.
      loop = None

    class TestCallback(callbacks.TrainingStepCallback):

      def __init__(self, loop):
        super().__init__(loop)
        ActualLoop.loop = loop

      def call_at(self, step):
        return False

      def on_step_begin(self, step):
        del step

      def on_step_end(self, step):
        del step

    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01)
    )
    expected_loop = training.Loop(
        model, [task], callbacks=[TestCallback]
    )
    self.assertIs(ActualLoop.loop, expected_loop)

  def test_calls_step_callbacks(self):
    """Runs a training loop, asserting that callbacks are called."""
    call_at_steps = [1, 3, 4]
    begin_steps = []
    end_steps = []
    test_case = self

    class TestCallback(callbacks.TrainingStepCallback):

      def call_at(self, step):
        return step in call_at_steps

      def on_step_begin(self, step):
        begin_steps.append(step)

      def on_step_end(self, step):
        # Assert that on_step_begin() was called before.
        test_case.assertIn(step, begin_steps)
        end_steps.append(step)

    model = tl.Serial(tl.Dense(1))
    task = training.TrainTask(
        _very_simple_data(), tl.L2Loss(), optimizers.SGD(.01)
    )
    loop = training.Loop(model, [task], callbacks=[TestCallback])
    loop.run(n_steps=5)

    # Assert that the callback has been called at the appropriate steps.
    self.assertEqual(begin_steps, call_at_steps)
    self.assertEqual(end_steps, call_at_steps)


def _very_simple_data(output_dim=1, input_dim=1):
  """"Returns stream of labeled data that maps small integers to constant pi."""
  inputs_batch = np.arange(8).reshape((8, 1))  # 8 items per batch
  inputs_batch = np.concatenate([inputs_batch] * input_dim, axis=1)
  targets_batch = np.pi * np.ones((8, output_dim))
  labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
  while True:
    yield labeled_batch


def _very_simple_transformer_data():
  """"Returns stream of labeled data that maps small integers to constant pi."""
  inputs_batch = np.ones((2, 2)).astype(np.int32)
  targets_batch = np.ones((2, 2, 8)).astype(np.int32)
  labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
  while True:
    yield labeled_batch


def _count_files(path):
  """Returns number of files in a given directory."""
  return len([filename for filename in os.listdir(path)
              if os.path.isfile(os.path.join(path, filename))])


if __name__ == '__main__':
  config.config_with_absl()
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Program to build a SubwordTextEncoder.

The flags --min_count and --corpus_max_lines will affect the size of the
vocabulary.  Try changing these flags until you get a vocabulary
of the size you want.

Example usage:

python trax/data/text_encoder_build_subword.py \
    --corpus_filepattern=$DATA_DIR/my_problem-train-* \
    --corpus_max_lines=12345 \
    --output_filename=$DATA_DIR/my_problem.subword_text_encoder \
    --logtostderr

"""

from absl import app
from absl import flags

from trax.data import text_encoder
from trax.data import tokenizer

flags.DEFINE_string('output_filename', '/tmp/my.subword_text_encoder',
                    'where to store the SubwordTextEncoder')
flags.DEFINE_string('corpus_filepattern', '',
                    'Corpus of one or more text files')
flags.DEFINE_string(
    'vocab_filepattern', '', 'One or more vocabulary files '
    '(one word per line as "word,count")')
flags.DEFINE_integer('min_count', 5, 'Minimum subtoken count in corpus')
flags.DEFINE_integer('corpus_max_lines', 10000,
                     'How many lines of corpus to read')
flags.DEFINE_integer('num_iterations', 4, 'Number of iterations')
flags.DEFINE_bool('split_on_newlines', True, 'Break corpus into lines.')

FLAGS = flags.FLAGS


def main(unused_argv):
  if FLAGS.corpus_filepattern and FLAGS.vocab_filepattern:
    raise ValueError(
        'Must only provide one of --corpus_filepattern or --vocab_filepattern')

  elif FLAGS.corpus_filepattern:
    token_counts = tokenizer.corpus_token_counts(
        FLAGS.corpus_filepattern,
        FLAGS.corpus_max_lines,
        split_on_newlines=FLAGS.split_on_newlines)

  elif FLAGS.vocab_filepattern:
    token_counts = tokenizer.vocab_token_counts(FLAGS.vocab_filepattern,
                                                FLAGS.corpus_max_lines)

  else:
    raise ValueError(
        'Must provide one of --corpus_filepattern or --vocab_filepattern')

  encoder = text_encoder.SubwordTextEncoder()
  encoder.build_from_token_counts(token_counts, FLAGS.min_count,
                                  FLAGS.num_iterations)
  encoder.store_to_file(FLAGS.output_filename)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax top level import."""

from trax import data
from trax import fastmath
from trax import layers
from trax import models
from trax import optimizers
from trax import shapes
from trax import supervised
from trax.supervised import lr_schedules as lr
from trax.trax2keras import AsKeras



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Simple net for playing Atari games using PPO."""

from trax import layers as tl


def _FrameStack(n_frames):
  """Stacks successive game frames along their last dimension."""
  # Input shape: (B, T, ..., C).
  # Output shape: (B, T, ..., C * n_frames).
  assert n_frames >= 1
  if n_frames == 1:
    return []  # No-op; just let the data flow through.
  return [
      # Create copies of input sequence, shift right by [0, ..., n_frames - 1]
      # frames, and concatenate along the channel dimension.
      tl.Branch(*map(_shift_right, range(n_frames))),
      tl.Concatenate(n_items=n_frames, axis=-1)
  ]


def _BytesToFloats():
  """Layer that converts unsigned bytes to floats."""
  return tl.Fn('BytesToFloats', lambda x: x / 255.0)


def AtariCnn(n_frames=4, hidden_sizes=(32, 32), output_size=128, mode='train'):
  """An Atari CNN."""
  del mode

  # TODO(jonni): Include link to paper?
  # Input shape: (B, T, H, W, C)
  # Output shape: (B, T, output_size)
  return tl.Serial(
      _BytesToFloats(),
      _FrameStack(n_frames=n_frames),  # (B, T, H, W, 4C)
      tl.Conv(hidden_sizes[0], (5, 5), (2, 2), 'SAME'),
      tl.Relu(),
      tl.Conv(hidden_sizes[1], (5, 5), (2, 2), 'SAME'),
      tl.Relu(),
      tl.Flatten(n_axes_to_keep=2),  # B, T and rest.
      tl.Dense(output_size),
      tl.Relu(),
  )


def AtariCnnBody(n_frames=4, hidden_sizes=(32, 64, 64),
                 output_size=512, mode='train',
                 kernel_initializer=None, padding='VALID'):
  """An Atari CNN."""
  del mode

  # TODO(jonni): Include link to paper?
  # Input shape: (B, T, H, W, C)
  # Output shape: (B, T, output_size)
  return tl.Serial(
      _BytesToFloats(),
      _FrameStack(n_frames=n_frames),  # (B, T, H, W, 4C)
      tl.Conv(hidden_sizes[0], (8, 8), (4, 4), padding=padding,
              kernel_initializer=kernel_initializer),
      tl.Relu(),
      tl.Conv(hidden_sizes[1], (4, 4), (2, 2), padding=padding,
              kernel_initializer=kernel_initializer),
      tl.Relu(),
      tl.Conv(hidden_sizes[2], (3, 3), (1, 1), padding=padding,
              kernel_initializer=kernel_initializer),
      tl.Relu(),
      tl.Flatten(n_axes_to_keep=2),  # B, T and rest.
      tl.Dense(output_size),
      tl.Relu(),
  )


def FrameStackMLP(n_frames=4, hidden_sizes=(64,), output_size=64,
                  mode='train'):
  """MLP operating on a fixed number of last frames."""
  del mode

  return tl.Serial(
      _FrameStack(n_frames=n_frames),
      [[tl.Dense(d_hidden), tl.Relu()] for d_hidden in hidden_sizes],
      tl.Dense(output_size),
  )


def _shift_right(n):  # pylint: disable=invalid-name
  return [tl.ShiftRight()] * n



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Test for memory usage in Reformer models.

This test is designed to run on TPUv3 hardware, processing 1 million tokens at a
time while just barely fitting within the 16 GB memory budget.
"""

import functools

from absl.testing import absltest

import jax
from jax import numpy as jnp
from jax import test_util  # pylint: disable=unused-import
from jax.config import config

import numpy as np

from trax import fastmath
from trax import layers as tl
from trax import shapes
from trax.models.reformer import reformer


class ReformerMemoryTest(absltest.TestCase):

  def test_reformer_lm_memory(self):
    lsh_self_attention = functools.partial(
        tl.LSHSelfAttention,
        attention_dropout=0.0,
        chunk_len=64,
        n_buckets=[128, 128],
        n_chunks_after=0,
        n_chunks_before=1,
        n_hashes=1,
        n_parallel_heads=1,
        predict_drop_len=128,
        predict_mem_len=1024,
    )
    timebin_self_attention = functools.partial(
        tl.SelfAttention,
        attention_dropout=0.05,
        chunk_len=64,
        n_chunks_before=1,
        n_parallel_heads=1,
    )

    model = reformer.ReformerLM(
        vocab_size=256,
        d_model=256,
        d_ff=512,
        d_attention_key=64,
        d_attention_value=64,
        n_layers=6,
        n_heads=2,
        dropout=0.05,
        max_len=1048576,
        attention_type=[timebin_self_attention, lsh_self_attention],
        pos_axial_shape=(1024, 1024),
        pos_d_axial_embs=(64, 192),
        ff_activation=tl.Relu,
        ff_use_sru=0,
        ff_chunk_size=131072,
        mode='train',
    )
    x = np.ones((1, 1048576)).astype(np.int32)
    weights, state = model.init(shapes.signature(x))

    @jax.jit
    def mock_training_step(x, weights, state, rng):
      def compute_mock_loss(weights):
        logits, new_state = model.pure_fn(x, weights, state, rng)
        loss = jnp.mean(logits[..., 0])
        return loss, (new_state, logits)
      gradients, (new_state, logits) = jax.grad(
          compute_mock_loss, has_aux=True)(weights)
      new_weights = fastmath.nested_map_multiarg(
          lambda w, g: w - 1e-4 * g, weights, gradients)
      return new_weights, new_state, logits

    weights, state, logits = mock_training_step(
        x, weights, state, jax.random.PRNGKey(0))
    self.assertEqual(logits.shape, (1, 1048576, 256))


if __name__ == '__main__':
  config.config_with_absl()
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.layers.attention."""

import functools
from absl.testing import absltest
import numpy as np

from trax import shapes
import trax.layers as tl
from trax.layers import test_utils


class AttentionTest(absltest.TestCase):

  def test_simple_call(self):
    layer = tl.CausalAttention(d_feature=4, n_heads=2)
    x = [np.array([[[2, 5, 3, 4],
                    [0, 1, 2, 3],
                    [0, 1, 2, 3],]]),
         np.array([[[[1, 0, 1]]]])]
    _, _ = layer.init(shapes.signature(x))

    y, mask = layer(x)
    self.assertEqual(y.shape, (1, 3, 4))
    self.assertEqual(mask.shape, (1, 1, 1, 3))

  def test_shift_right(self):
    # Test shifts right on axis=1
    layer = tl.ShiftRight()
    x = np.array([[[9, 9, 9],
                   [8, 8, 8],
                   [7, 7, 7],
                   [6, 6, 6]],
                  [[99, 98, 97],
                   [96, 95, 94],
                   [93, 92, 91],
                   [90, 89, 88]]])
    y = layer(x)
    self.assertEqual(x.shape, y.shape)
    self.assertEqual(tl.to_list(y), [[[0, 0, 0],
                                      [9, 9, 9],
                                      [8, 8, 8],
                                      [7, 7, 7]],
                                     [[0, 0, 0],
                                      [99, 98, 97],
                                      [96, 95, 94],
                                      [93, 92, 91]]])

  def test_shift_right_float(self):
    layer = tl.ShiftRight()
    x = np.array([[[9, 9, 9],
                   [8, 8, 8],
                   [7, 7, 7],
                   [6, 6, 6]],
                  [[99, 98, 97],
                   [96, 95, 94],
                   [93, 92, 91],
                   [90, 89, 88]]]).astype(np.float32)
    x /= 2.0
    self.assertEqual(x.dtype, np.float32)

    y = layer(x)
    self.assertEqual(y.dtype, np.float32)
    self.assertEqual(tl.to_list(y), [[[0.0, 0.0, 0.0],
                                      [4.5, 4.5, 4.5],
                                      [4.0, 4.0, 4.0],
                                      [3.5, 3.5, 3.5]],
                                     [[0.0, 0.0, 0.0],
                                      [49.5, 49.0, 48.5],
                                      [48.0, 47.5, 47.0],
                                      [46.5, 46.0, 45.5]]])

  def test_padding_mask(self):
    layer = tl.PaddingMask()
    x = np.array([
        [1., 2., 3., 4., 0.],
        [1., 2., 3., 0., 0.],
        [1., 2., 0., 0., 0.],
    ])
    y = layer(x)
    self.assertEqual(x.shape, (3, 5))
    self.assertEqual(y.shape, (3, 1, 1, 5))
    np.testing.assert_equal(y, [[[[True, True, True, True, False]]],
                                [[[True, True, True, False, False]]],
                                [[[True, True, False, False, False]]]])


class CausalAttentionTest(absltest.TestCase):

  def test_simple_call(self):
    layer = tl.CausalAttention(d_feature=4, n_heads=2)
    x = np.array([[[2, 5, 3, 4],
                   [0, 1, 2, 3],
                   [0, 1, 2, 3],]])
    _, _ = layer.init(shapes.signature(x))

    y = layer(x)
    self.assertEqual(y.shape, (1, 3, 4))

  def test_deterministic_eval(self):
    d_model = 32
    seq_len = 3
    x_shape = (1, seq_len, d_model)
    inp = np.ones(x_shape).astype(np.float32)

    model_fn = functools.partial(
        tl.CausalAttention,
        d_feature=d_model,
        n_heads=4,
        )

    test_utils.test_eval_is_deterministic(inp, model_fn)

  def test_predict_equals_eval(self):
    d_model = 32
    seq_len = 10
    x_shape = (1, seq_len, d_model)
    inp = np.ones(x_shape).astype(np.float32)

    model_fn = functools.partial(
        tl.CausalAttention,
        d_feature=d_model,
        n_heads=4,
        )

    test_utils.test_eval_equals_predict(inp, model_fn)


class PositionalEncodingTest(absltest.TestCase):

  def test_simple_call(self):
    layer = tl.PositionalEncoding(max_len=8)
    x = np.array([[[2.0, 3.0, 4.0, 5.0], [1.0, 2.0, 3.0, 4.0]]])
    layer.init(shapes.signature(x))
    y = layer(x)
    self.assertEqual(y.shape, (1, 2, 4))

  def test_predict(self):
    layer = tl.PositionalEncoding(max_len=8)
    x = np.array([[[2.0, 3.0], [1.0, 2.0], [0.0, 1.0], [3.0, 4.0]]])
    self.assertEqual(x.shape, (1, 4, 2))
    layer.init(shapes.signature(x))
    y = layer(x)
    self.assertEqual(y.shape, (1, 4, 2))
    layer = tl.PositionalEncoding(max_len=8, mode='predict')
    layer.init(shapes.signature(x[:, :1, :]))
    y0 = layer(x[:, :1, :])   # just the first token
    self.assertEqual(y0.shape, (1, 1, 2))
    self.assertTrue(np.array_equal(y0, y[:, :1, :]))
    y1 = layer(x[:, 1:3, :])  # now the next 2 tokens
    self.assertEqual(y1.shape, (1, 2, 2))
    self.assertTrue(np.array_equal(y1, y[:, 1:3, :]))
    y2 = layer(x[:, 3:4, :])  # final one token
    self.assertEqual(y2.shape, (1, 1, 2))
    self.assertTrue(np.array_equal(y2, y[:, 3:4, :]))

  def test_predict_equals_eval(self):
    x = np.array([[[2.0, 3.0], [1.0, 2.0], [0.0, 1.0], [3.0, 4.0]]])
    self.assertEqual(x.shape, (1, 4, 2))

    layer_eval = tl.PositionalEncoding(max_len=8, d_feature=4, mode='eval')
    layer_eval.init(shapes.signature(x))

    output_eval = layer_eval(x)

    layer_predict = tl.PositionalEncoding(max_len=8, d_feature=4,
                                          mode='predict')
    layer_predict.init(shapes.signature(x))
    layer_predict.weights = layer_eval.weights

    output_predict = layer_predict(x)
    self.assertTrue(np.array_equal(output_eval, output_predict))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NumPy like wrapper for Tensorflow."""

# pylint: disable=wildcard-import
# pylint: disable=g-import-not-at-top
# pylint: disable=g-direct-tensorflow-import

try:
  # Note that this import will work in tf-nightly and TF versions 2.4 and
  # higher.
  from tensorflow.experimental.numpy import *
  # TODO(agarwal): get rid of following imports.
  from tensorflow.experimental.numpy import random
  from tensorflow import bfloat16
  import numpy as onp
  from tensorflow.python.ops.numpy_ops.np_dtypes import canonicalize_dtype
  from tensorflow.python.ops.numpy_ops.np_dtypes import default_float_type
  from tensorflow.python.ops.numpy_ops.np_dtypes import is_allow_float64
  from tensorflow.python.ops.numpy_ops.np_dtypes import set_allow_float64

  random.DEFAULT_RANDN_DTYPE = onp.float32
except ImportError:
  try:
    # Note that this import will work in TF 2.3 and higher.
    from tensorflow.python.ops.numpy_ops import *
    from tensorflow import bfloat16

  except ImportError:
    # Note that this fallback will be needed for TF 2.2.
    from tensorflow import newaxis

    from trax.tf_numpy.numpy_impl import random

    # pylint: disable=wildcard-import
    from trax.tf_numpy.numpy_impl.array_ops import *
    from trax.tf_numpy.numpy_impl.arrays import *
    from trax.tf_numpy.numpy_impl.dtypes import *
    from trax.tf_numpy.numpy_impl.math_ops import *
    from trax.tf_numpy.numpy_impl.utils import finfo
    from trax.tf_numpy.numpy_impl.utils import promote_types
    from trax.tf_numpy.numpy_impl.utils import result_type
    # pylint: enable=wildcard-import

    max = amax  # pylint: disable=redefined-builtin,undefined-variable
    min = amin  # pylint: disable=redefined-builtin,undefined-variable
    round = around  # pylint: disable=redefined-builtin,undefined-variable

try:
  from tensorflow.python.ops.numpy_ops.np_config import enable_numpy_behavior
  # TODO(b/171429739): This should be moved to every individual file/test.
  enable_numpy_behavior()

except ImportError:
  pass



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""SM3 optimizer class."""

import enum

from trax.fastmath import numpy as jnp
from trax.optimizers import base as opt_base


class MomentumType(enum.IntEnum):
  EMA = 1
  HEAVY_BALL = 2
  NESTEROV = 3


class SM3(opt_base.Optimizer):
  """SM3 optimizer, as described in https://arxiv.org/abs/1901.11150."""

  def __init__(self,
               learning_rate=0.01,
               momentum=0.9,
               second_moment_averaging=1.0,
               weight_decay=0.0,
               momentum_type=MomentumType.EMA):  # pylint: disable=useless-super-delegation
    """Create the SM3 optimizer.

    Memory-Efficient Adaptive Optimization.
    https://arxiv.org/abs/1901.11150

    Args:
      learning_rate: a postitive scalar value for the initial learning rate.
      momentum: optional, a positive scalar value for momentum
      second_moment_averaging: averaging of second moments (if 1.0, adds from
        begining of time like AdaGrad).
      weight_decay: Weight decay for regularizing the model.
      momentum_type: Nestrov, Heavy-Ball or EMA (Default).

    """
    self._has_momentum = momentum > 0.0
    self._momentum_type = momentum_type
    self._graft = second_moment_averaging != 1.0
    super().__init__(
        learning_rate=learning_rate,
        momentum=momentum,
        second_moment_averaging=second_moment_averaging,
        weight_decay=weight_decay,
    )

  def init(self, w):
    momentum = []
    if self._has_momentum:
      momentum = jnp.zeros_like(w)
    v1s = [jnp.zeros(sz, dtype=w.dtype) for sz in w.shape]
    v2s = []
    if self._graft:
      v2s = [jnp.zeros(sz, dtype=w.dtype) for sz in w.shape]
    return (momentum, v1s, v2s)

  def _momentum_update(self, g, m, beta1):
    """Handle various types of momentum."""
    if self._momentum_type == MomentumType.EMA:
      m = (1 - beta1) * g + beta1 * m
      update = m
    elif self._momentum_type == MomentumType.HEAVY_BALL:
      m = g + beta1 * m
      update = m
    elif self._momentum_type == MomentumType.NESTEROV:
      m = g + beta1 * m
      nesterov_m = g + beta1 * m
      update = nesterov_m
    else:
      assert False, 'Unknown momentum_type.'
    return m, update

  def _update_diagonal(self, g, w, m, v1, v2, opt_params):
    learning_rate = opt_params['learning_rate']
    beta2 = opt_params['second_moment_averaging']
    weight_decay = opt_params['weight_decay']

    is_beta2_1 = (beta2 == 1).astype(g.dtype)
    one_minus_beta2_except1 = is_beta2_1  + (1.0 - beta2) * (1.0 - is_beta2_1)
    v1[0] = beta2 * v1[0] + one_minus_beta2_except1 * g * g

    preconditioner = jnp.where(v1[0] > 0, 1.0 / (jnp.sqrt(v1[0]) + 1e-16),
                               jnp.zeros_like(v1[0]))

    pg = preconditioner * g
    if self._graft:
      v2[0] += g * g
      preconditioner_graft = jnp.where(
          v2[0] > 0, 1.0 / (jnp.sqrt(v2[0]) + 1e-16), jnp.zeros_like(v2[0]))
      pg_graft = preconditioner_graft * g
      pg_norm = jnp.linalg.norm(pg)
      pg_graft_norm = jnp.linalg.norm(pg_graft)
      pg = pg * (pg_graft_norm/(pg_norm + 1e-16))

    pg = pg + w * weight_decay

    if self._has_momentum:
      m, update = self._momentum_update(pg, m, opt_params['momentum'])
    else:
      update = pg

    w = w - (update * learning_rate).astype(w.dtype)
    return w, (m, v1, v2)

  def _expanded_shape(self, shape, axis):
    # Replaces a `shape` of [M, N, K] with 1 in all dimensions except for i.
    # For eg: i = 1 returns [1, N, 1].
    rank = len(shape)
    return [1] * axis + [shape[axis]] + [1] * (rank - axis - 1)

  def _minimum(self, tensor_list):
    minimum = tensor_list[0]
    for i in range(1, len(tensor_list)):
      minimum = jnp.minimum(minimum, tensor_list[i])
    return minimum

  def _update_sketched(self, g, w, m, v1, v2, opt_params):
    """Update for higher-rank parameters."""
    learning_rate = opt_params['learning_rate']
    momentum = opt_params['momentum']
    beta2 = opt_params['second_moment_averaging']
    weight_decay = opt_params['weight_decay']

    shape = w.shape
    rank = len(shape)
    reshaped_accumulators = [jnp.reshape(v1[i], self._expanded_shape(shape, i))
                             for i in range(rank)]
    acc = self._minimum(reshaped_accumulators)

    is_beta2_1 = (beta2 == 1).astype(g.dtype)
    one_minus_beta2_except1 = is_beta2_1  + (1.0 - beta2) * (1.0 - is_beta2_1)
    acc = beta2 * acc + one_minus_beta2_except1 * g * g

    preconditioner = jnp.where(acc > 0.0, 1.0 / (jnp.sqrt(acc) + 1e-16),
                               jnp.zeros_like(acc))
    pg = g * preconditioner
    if self._graft:
      v2_acc = self._minimum([
          jnp.reshape(v2[i], self._expanded_shape(shape, i))
          for i in range(rank)
      ])
      v2_acc = v2_acc + g * g
      preconditioner_graft = jnp.where(v2_acc > 0.0,
                                       1.0 / (jnp.sqrt(v2_acc) + 1e-16),
                                       jnp.zeros_like(v2_acc))
      pg_graft = preconditioner_graft * g
      pg_norm = jnp.linalg.norm(pg)
      pg_graft_norm = jnp.linalg.norm(pg_graft)
      pg = pg * (pg_graft_norm/(pg_norm + 1e-16))

    pg = pg + w * weight_decay

    if self._has_momentum:
      m, update = self._momentum_update(pg, m, momentum)
    else:
      update = pg

    w = w - (learning_rate * update).astype(w.dtype)
    for i in range(len(v1)):
      axes = list(range(int(i))) + list(range(int(i) + 1, rank))
      dim_accumulator = jnp.amax(acc, axis=axes)
      v1[i] = dim_accumulator

    if self._graft:
      for i in range(len(v2)):
        axes = list(range(int(i))) + list(range(int(i) + 1, rank))
        dim_accumulator = jnp.amax(v2_acc, axis=axes)
        v2[i] = dim_accumulator
    return w, (m, v1, v2)

  def update(self, step, g, w, slots, opt_params):
    del step
    m, v1, v2 = slots
    rank = len(w.shape)
    if rank > 1:
      return self._update_sketched(g, w, m, v1, v2, opt_params)
    else:
      return self._update_diagonal(g, w, m, v1, v2, opt_params)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax history."""
import collections
import copy

from absl import logging
import six


class History:
  """History of metrics.

  History contains the metrics recorded during training and evaluation.
  Save data with history.append and get a sequence of data by calling
  history.get.

  For example:
  history.append('train', 'metrics/accuracy', 1, 0.04)
  history.append('train', 'metrics/accuracy', 1000, 0.31)
  history.get('train', 'metrics/accuracy')
  # returns [(1, 0.04), (1000, 0.31)]
  """

  def __init__(self):
    # Structure is
    # values = {
    #   'mode1': {
    #     'metric1': [val1, val2],
    #     ...
    #   },
    #   'mode2': ...
    # }
    self._values = {}

  @classmethod
  def from_dict(cls, json_object):
    """Constructs a `History` from a Python dictionary of parameters."""
    history = History()
    for (key, value) in six.iteritems(json_object):
      history.__dict__[key] = value
    return history

  def to_dict(self):
    """Serializes this instance to a Python dictionary."""
    output = copy.deepcopy(self.__dict__)
    return output

  def append(self, mode, metric, step, value):
    """Append (step, value) pair to history for the given mode and metric."""
    if mode not in self._values:
      self._values[mode] = collections.defaultdict(list)
    self._values[mode][metric].append((step, value))

  def get(self, mode, metric):
    """Get the history for the given metric and mode."""
    if mode not in self._values:
      logging.info('Metric %s not found for mode %s', metric, mode)
      return []
    return list(self._values[mode][metric])

  @property
  def modes(self):
    """Current tracked modes."""
    return sorted(list(self._values.keys()))

  def metrics_for_mode(self, mode):
    """Metrics available for a given mode."""
    if mode not in self._values:
      logging.info('Mode %s not found', mode)
      return []
    return sorted(list(self._values[mode].keys()))

  def __str__(self):
    return str(self._values)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for combinator layers."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from trax import fastmath
from trax import shapes
import trax.layers as tl


def DivideBy(val):  # pylint: disable=invalid-name
  """Returns a simple division layer with n_in == 1 and n_out == 1."""
  return tl.Fn('DivideBy', lambda x: x / val)


def ReturnConst(val):  # pylint: disable=invalid-name
  """Returns a simple const layer with n_in == 0 and n_out == 1."""
  return tl.Fn('ReturnConst', lambda: val)


def SmallerThan(val):  # pylint: disable=invalid-name
  """Checks if the input is smaller than certain value."""
  return tl.Fn('SmallerThan', lambda x: x < val)


# TODO(jonni): Consider a more generic home for this utiliity function.
def as_list(outputs):
  """Converts layer outputs to a nested list, for easier equality testing.

  Args:
    outputs: A tensor or tuple/list of tensors coming from the forward
        application of a layer. Each tensor is NumPy ndarray-like, which
        complicates simple equality testing (e.g., via `assertEquals`):
        such tensors require equality testing to use either `all` (all
        elements match) or `any` (at least one element matches), which is not
        directly supported in absltest.

  Returns:
    A nested list structure containing all the output values, but now directly
    testable using `assertEquals`.
  """
  if isinstance(outputs, (list, tuple)):
    return [as_list(y) for y in outputs]
  else:
    return outputs.tolist()


class SerialTest(absltest.TestCase):

  def test_none_is_no_op(self):
    layer = tl.Serial(None)
    xs = [np.array([1, 2, 3, 4]),
          np.array([10, 20, 30])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[1, 2, 3, 4],
                                   [10, 20, 30]])

  def test_empty_list_is_no_op(self):
    layer = tl.Serial([])
    xs = [np.array([1, 2, 3, 4]),
          np.array([10, 20, 30])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[1, 2, 3, 4],
                                   [10, 20, 30]])

  def test_one_in_one_out(self):
    layer = tl.Serial(DivideBy(3))
    x = np.array([3, 6, 9, 12])
    y = layer(x)
    self.assertEqual(as_list(y), [1, 2, 3, 4])

  def test_zero_in_one_out(self):
    layer = tl.Serial(ReturnConst(np.array([3, 4, 5, 6])))
    y = layer(())
    self.assertEqual(as_list(y), [3, 4, 5, 6])

  def test_one_in_two_out(self):
    layer = tl.Serial(DivideBy(3),
                      ReturnConst(np.array([3, 4, 5, 6])))
    x = np.array([3, 6, 9, 12])
    y = layer(x)
    self.assertEqual(as_list(y), [[3, 4, 5, 6],
                                  [1, 2, 3, 4]])

  def test_const_div(self):
    layer = tl.Serial(ReturnConst(np.array([3, 6, 9, 12])),
                      DivideBy(3))
    y = layer(())
    self.assertEqual(as_list(y), [1, 2, 3, 4])

  def test_div_div(self):
    layer = tl.Serial(DivideBy(2.0), DivideBy(5.0))
    x = np.array([10, 20, 30])
    y = layer(x)
    self.assertEqual(as_list(y), [1, 2, 3])

  def test_dup_dup(self):
    layer = tl.Serial(tl.Dup(), tl.Dup())
    x = np.array([1, 2, 3])
    ys = layer(x)
    self.assertEqual(as_list(ys), [[1, 2, 3],
                                   [1, 2, 3],
                                   [1, 2, 3]])

  def test_default_name(self):
    layer = tl.Serial(tl.Dup(), tl.Dup())
    self.assertIn('Serial', str(layer))

  def test_custom_name(self):
    layer = tl.Serial(tl.Dup(), tl.Dup(), name='Branch')
    self.assertIn('Branch', str(layer))

  def test_weights(self):
    model = tl.Serial(tl.Dense(4), tl.Dense(5), tl.Dense(7))
    self.assertIsInstance(model.weights, tuple)
    self.assertLen(model.weights, 3)

  def test_flat_weights_and_state(self):
    model = tl.Serial(tl.Dup(), tl.Dense(5), tl.Serial(tl.Dense(7), tl.Dup()))
    sample_input_signature = shapes.signature(np.zeros((2, 3)))
    model.init(sample_input_signature)
    flat_weights, flat_state = tl.flatten_weights_and_state(
        model.weights, model.state)
    # Model has 2 pairs of trainable weights: (w, b) for the 2 dense layers.
    # So after making them flat, there are 4 trainable weights.
    self.assertLen(flat_weights, 4)
    self.assertEmpty(flat_state)
    model2 = tl.Serial(tl.Dense(5), tl.Dup(), tl.Dense(7))
    sig = model2.weights_and_state_signature(sample_input_signature)
    weights2, state2 = tl.unflatten_weights_and_state(
        flat_weights, flat_state, sig)
    model2.weights = weights2
    model2.state = state2
    self.assertLen(model2.weights, 3)
    self.assertEqual(model.weights[1], model2.weights[0])
    self.assertEqual(model.weights[2][0], model2.weights[2])

  def test_flat_weights_and_state_shared(self):
    shared = tl.Dense(5)
    model = tl.Serial(tl.Dense(5), shared, tl.Serial(shared, tl.Dup()))
    sample_input_signature = shapes.signature(np.zeros((2, 3)))
    model.init(sample_input_signature)
    flat_weights, flat_state = tl.flatten_weights_and_state(
        model.weights, model.state)
    # Model has 2 pairs of trainable weights: (w, b) for the 2 dense layers.
    # So after making them flat, there are 4 trainable weights.
    self.assertLen(flat_weights, 4)
    self.assertEmpty(flat_state)
    model2 = tl.Serial(tl.Dense(5), tl.Dup(), tl.Dense(5))
    sig = model2.weights_and_state_signature(sample_input_signature)
    weights2, state2 = tl.unflatten_weights_and_state(
        flat_weights, flat_state, sig)
    model2.weights = weights2
    model2.state = state2
    self.assertLen(model2.weights, 3)
    self.assertEqual(model.weights[0], model2.weights[0])
    self.assertEqual(model.weights[1], model2.weights[2])

  def test_assign_sublayer_weights(self):
    layer = tl.Dense(5, use_bias=False)
    model = tl.Serial(tl.Serial(layer, tl.Dense(6)), tl.Dense(7))
    sample_input = np.array([1, 2, 3, 4, 5])
    weights, _ = model.init(shapes.signature(sample_input))
    new_layer_weights = np.random.uniform(weights[0][0].shape)
    layer.weights = new_layer_weights
    self.assertIs(model.weights[0][0], new_layer_weights)

  def test_shared_weights(self):
    layer = tl.Dense(5)
    model = tl.Serial(layer, layer)
    sample_input = np.array([1, 2, 3, 4, 5])
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1], tl.GET_WEIGHTS_FROM_CACHE)

  def test_shared_weights_nested(self):
    layer = tl.Dense(5)
    model = tl.Serial(layer, tl.Serial(layer))
    sample_input = np.array([1, 2, 3, 4, 5])
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1][0], tl.GET_WEIGHTS_FROM_CACHE)

  def test_shared_weights_double_nested(self):
    layer = tl.Dense(5)
    model = tl.Serial(tl.Serial(layer), tl.Serial(layer))
    sample_input = np.array([1, 2, 3, 4, 5])
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1][0], tl.GET_WEIGHTS_FROM_CACHE)

  def test_shared_weights_for_shared_serial(self):
    layer = tl.Serial(tl.Dense(5), tl.Dense(5))
    model = tl.Serial(layer, layer)
    sample_input = np.array([1, 2, 3, 4, 5])
    # Init gives weights reflecting weight sharing.
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIsNot(weights[0], tl.GET_WEIGHTS_FROM_CACHE)
    self.assertIs(weights[1], tl.GET_WEIGHTS_FROM_CACHE)
    # Forward pass runs successfully.
    y = model(sample_input)
    self.assertEqual(y.shape, (5,))

  def test_state(self):
    model = tl.Serial(tl.Dense(4), tl.Dense(5), tl.Dense(7))
    self.assertIsInstance(model.state, tuple)
    self.assertLen(model.state, 3)

  def test_set_rng_recurse_two_levels(self):
    dense_00 = tl.Dense(2)
    dense_01 = tl.Dense(2)
    dense_10 = tl.Dense(2)
    dense_11 = tl.Dense(2)
    layer = tl.Serial(
        tl.Serial(dense_00, dense_01),
        tl.Serial(dense_10, dense_11),
    )
    input_signature = shapes.ShapeDtype((1, 2))

    _, _ = layer.init(input_signature)
    weights = layer.weights
    dense_00_w, dense_00_b = weights[0][0]
    dense_01_w, dense_01_b = weights[0][1]
    dense_10_w, dense_10_b = weights[1][0]
    dense_11_w, dense_11_b = weights[1][1]

    # Setting rng's recursively during init should yield differing weights.
    self.assertFalse(np.array_equal(dense_00_w, dense_01_w))
    self.assertFalse(np.array_equal(dense_00_b, dense_01_b))
    self.assertFalse(np.array_equal(dense_10_w, dense_11_w))
    self.assertFalse(np.array_equal(dense_10_b, dense_11_b))


class ParallelTest(absltest.TestCase):

  def test_dup_dup(self):
    layer = tl.Parallel(tl.Dup(), tl.Dup())
    xs = [np.array([1, 2, 3]),
          np.array([10, 20])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[1, 2, 3],
                                   [1, 2, 3],
                                   [10, 20],
                                   [10, 20]])

  def test_div_div(self):
    layer = tl.Parallel(DivideBy(0.5), DivideBy(3.0))
    xs = [np.array([1, 2, 3]),
          np.array([30, 60])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[2, 4, 6],
                                   [10, 20]])

  def test_two_no_ops(self):
    layer = tl.Parallel([], None)
    xs = [np.array([1, 2, 3]),
          np.array([10, 20])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[1, 2, 3],
                                   [10, 20]])

  def test_default_name(self):
    layer = tl.Parallel(tl.Dup(), tl.Dup())
    self.assertIn('Parallel', str(layer))

  def test_custom_name(self):
    layer = tl.Parallel(tl.Dup(), tl.Dup(), name='DupDup')
    self.assertIn('DupDup', str(layer))

  def test_weights(self):
    model = tl.Parallel(tl.Dense(3), tl.Dense(5))
    self.assertIsInstance(model.weights, tuple)
    self.assertLen(model.weights, 2)

  def test_shared_weights(self):
    layer = tl.Dense(5)
    model = tl.Parallel(layer, layer)
    sample_input = (np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]))
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1], tl.GET_WEIGHTS_FROM_CACHE)

  def test_shared_weights_nested(self):
    layer = tl.Dense(5)
    model = tl.Parallel([layer, tl.Dense(2)],
                        [layer, tl.Dense(2)])
    sample_input = (np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]))
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1][0], tl.GET_WEIGHTS_FROM_CACHE)

  def test_shared_weights_for_shared_parallel(self):
    layer = tl.Parallel(tl.Dense(5), tl.Dense(7))
    model = tl.Parallel(layer, layer)
    sample_input = [
        np.array([1, 2, 3]),
        np.array([10, 20, 30]),
        np.array([100, 200, 300]),
        np.array([1000, 2000, 3000]),
    ]
    # Init gives weights reflecting weight sharing.
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIsNot(weights[0], tl.GET_WEIGHTS_FROM_CACHE)
    self.assertIs(weights[1], tl.GET_WEIGHTS_FROM_CACHE)
    # Forward pass runs successfully.
    y0, y1, y2, y3 = model(sample_input)
    self.assertEqual(y0.shape, (5,))
    self.assertEqual(y1.shape, (7,))
    self.assertEqual(y2.shape, (5,))
    self.assertEqual(y3.shape, (7,))

  def test_state(self):
    model = tl.Parallel(tl.Dense(3), tl.Dense(5))
    self.assertIsInstance(model.state, tuple)
    self.assertLen(model.state, 2)


class ConcatenateTest(absltest.TestCase):

  def test_n_in_n_out(self):
    layer = tl.Concatenate()
    self.assertEqual(layer.n_in, 2)
    self.assertEqual(layer.n_out, 1)

  def test_with_defaults(self):
    layer = tl.Concatenate()  # Default n_items=2, axis=-1
    xs = [np.array([[1, 2, 3],
                    [4, 5, 6]]),
          np.array([[10, 20, 30],
                    [40, 50, 60]])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[1, 2, 3, 10, 20, 30],
                                   [4, 5, 6, 40, 50, 60]])

  def test_axis_0(self):
    layer = tl.Concatenate(axis=0)
    xs = [np.array([[1, 2, 3],
                    [4, 5, 6]]),
          np.array([[10, 20, 30],
                    [40, 50, 60]])]
    y = layer(xs)
    self.assertEqual(as_list(y), [[1, 2, 3],
                                  [4, 5, 6],
                                  [10, 20, 30],
                                  [40, 50, 60]])

  def test_axis_1(self):
    layer = tl.Concatenate(axis=1)
    xs = [np.array([[1, 2, 3],
                    [4, 5, 6]]),
          np.array([[10, 20, 30],
                    [40, 50, 60]])]
    y = layer(xs)
    self.assertEqual(as_list(y), [[1, 2, 3, 10, 20, 30],
                                  [4, 5, 6, 40, 50, 60]])

  def test_n_items_is_not_default(self):
    layer = tl.Concatenate(n_items=3)
    xs = [np.array([[1, 2, 3],
                    [4, 5, 6]]),
          np.array([[10, 20, 30],
                    [40, 50, 60]]),
          np.array([[100, 200, 300],
                    [400, 500, 600]])]
    y = layer(xs)
    self.assertEqual(y.shape, (2, 9))
    self.assertEqual(as_list(y), [[1, 2, 3, 10, 20, 30, 100, 200, 300],
                                  [4, 5, 6, 40, 50, 60, 400, 500, 600]])

  def test_repr(self):
    layer = tl.Concatenate()
    self.assertEqual(repr(layer), 'Concatenate_in2')

    layer = tl.Concatenate(axis=0)
    self.assertEqual(repr(layer), 'Concatenate_axis0_in2')

    layer = tl.Concatenate(axis=1)
    self.assertEqual(repr(layer), 'Concatenate_axis1_in2')

    layer = tl.Concatenate(n_items=3)
    self.assertEqual(repr(layer), 'Concatenate_in3')


class BranchTest(absltest.TestCase):

  def test_noop_dup(self):
    layer = tl.Branch([], tl.Dup())
    x = np.array([1, 2, 3])
    ys = layer(x)
    self.assertEqual(as_list(ys), [[1, 2, 3],
                                   [1, 2, 3],
                                   [1, 2, 3]])

  def test_add_div(self):
    layer = tl.Branch(tl.Add(), DivideBy(0.5))
    xs = [np.array([1, 2, 3]),
          np.array([10, 20, 30])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[11, 22, 33],
                                   [2, 4, 6]])

  def test_one_sublayer(self):
    layer = tl.Branch(DivideBy(0.5))
    x = np.array([1, 2, 3])
    ys = layer(x)
    self.assertEqual(as_list(ys), [2, 4, 6])

  def test_default_name(self):
    layer = tl.Branch(tl.Add(), DivideBy(0.5))
    self.assertIn('Branch', str(layer))

  def test_printing_sublayers(self):
    layer = tl.Branch(tl.Add(), tl.Add())
    expected_result = 'Branch_in2_out2[\n  Add_in2\n  Add_in2\n]'
    self.assertEqual(expected_result, str(layer))


class SelectTest(absltest.TestCase):

  def test_computes_n_in(self):
    layer = tl.Select([0, 0])
    self.assertEqual(layer.n_in, 1)

    layer = tl.Select([1, 0])
    self.assertEqual(layer.n_in, 2)

    layer = tl.Select([2])
    self.assertEqual(layer.n_in, 3)

  def test_given_n_in(self):
    layer = tl.Select([0], n_in=2)
    self.assertEqual(layer.n_in, 2)

    layer = tl.Select([0], n_in=3)
    self.assertEqual(layer.n_in, 3)

  def test_first_of_3(self):
    layer = tl.Select([0], n_in=3)
    xs = [np.array([1, 2, 3]),
          np.array([10, 20]),
          np.array([100])]
    y = layer(xs)
    self.assertEqual(as_list(y), [1, 2, 3])

  def test_second_of_3(self):
    layer = tl.Select([1], n_in=3)
    xs = [np.array([1, 2, 3]),
          np.array([10, 20]),
          np.array([100])]
    y = layer(xs)
    self.assertEqual(as_list(y), [10, 20])


class DropTest(absltest.TestCase):

  def test_drop(self):
    layer = tl.Drop()
    x = np.array([1, 2, 3])
    y = layer(x)
    self.assertEqual(as_list(y), [])


class SwapTest(absltest.TestCase):

  def test_swap(self):
    layer = tl.Swap()
    xs = [np.array([1, 2, 3]),
          np.array([10, 20, 30])]
    ys = layer(xs)
    self.assertEqual(as_list(ys), [[10, 20, 30],
                                   [1, 2, 3]])


class ChunkTest(absltest.TestCase):

  def test_chunk(self):
    layer = tl.Dense(4)
    x = np.array([[1, 2, 3], [4, 5, 6]])
    layer.init(x)
    y = layer(x)
    z = tl.Chunk(layer, 1)(x)
    self.assertLess(np.sum((y - z)**2), 1e-5)  # y == z upto numerics

  def test_chunk_uneven_numbers(self):
    layer = tl.Dense(4)
    x = np.array([[1, 2, 3], [4, 5, 6]])
    layer.init(x)
    y = layer(x)
    z = tl.Chunk(layer, 3)(x)  # By default it should just pass
    self.assertLess(np.sum((y - z)**2), 1e-5)  # y == z upto numerics
    chunk_with_test = tl.Chunk(layer, 3, pass_unchunkable=False)
    self.assertRaises(tl.LayerError, lambda: chunk_with_test(x))


class SerialWithSideOutputsTest(absltest.TestCase):

  def test_serial_with_side_outputs_div_div(self):
    def some_layer():
      return tl.Parallel(DivideBy(2.0), DivideBy(5.0))
    layer = tl.SerialWithSideOutputs([some_layer(), some_layer()])
    xs = (np.array([1, 2, 3]),
          np.array([10, 20, 30, 40, 50]),
          np.array([100, 200]))
    ys = layer(xs)
    output_shapes = [y.shape for y in ys]
    self.assertEqual(output_shapes, [(3,), (5,), (2,)])


BACKENDS = [fastmath.Backend.JAX]


@parameterized.named_parameters(
    ('_' + b.value, b) for b in BACKENDS)
class ScanTest(parameterized.TestCase):

  def _AddWithCarry(self):  # pylint: disable=invalid-name
    del self
    def f(x, carry):
      res = x + carry
      return res, res  # output and carry are the same
    return tl.Fn('AddWithCarry', f, n_out=2)

  def test_default_axis(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.Scan(self._AddWithCarry())
      xs = [
          np.array([[0, 1, 2, 3],
                    [0, 10, 20, 30],
                    [0, 100, 200, 300]]),
          np.array([9000, 8000, 7000, 6000])
      ]
      ys = layer(xs)
      self.assertEqual(as_list(ys),
                       [[[9000, 8001, 7002, 6003],
                         [9000, 8011, 7022, 6033],
                         [9000, 8111, 7222, 6333]
                        ],
                        [9000, 8111, 7222, 6333]
                       ])

  def test_axis_1(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.Scan(self._AddWithCarry(), axis=1)
      xs = [
          np.array([[0, 1, 2, 3],
                    [0, 10, 20, 30],
                    [0, 100, 200, 300]]),
          np.array([9000,
                    8000,
                    7000])
      ]
      ys = layer(xs)
      self.assertEqual(as_list(ys),
                       [[[9000, 9001, 9003, 9006],
                         [8000, 8010, 8030, 8060],
                         [7000, 7100, 7300, 7600]
                        ],
                        [9006,
                         8060,
                         7600]
                       ])

  def test_predict(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.Scan(self._AddWithCarry(), axis=1, mode='predict')
      xs = [np.array([[0, 1, 2]]),
            np.array([90])]
      ys = layer(xs)
      self.assertEqual(as_list(ys),
                       [[[90, 91, 93]],
                        [93]])
      xs = [np.array([[3, 4]]),
            np.array([90])]
      ys = layer(xs)
      self.assertEqual(as_list(ys),
                       [[[96, 100]],
                        [100]])

  def test_multi_input(self, backend):
    def _MultiInputFn():  # pylint: disable=invalid-name
      def f(a, b, carry):
        return a + b, b, carry + 1
      return tl.Fn('MultiInputFn', f, n_out=2)

    with fastmath.use_backend(backend):
      layer = tl.Scan(_MultiInputFn(), axis=1)
      xs = [
          np.array([[0, 1, 2],
                    [0, 10, 20]]),
          np.array([[4, 5, 6],
                    [40, 50, 60]]),
          np.array([9000,
                    8000])
      ]
      ys = layer(xs)
      self.assertEqual(as_list(ys),
                       [[[4, 6, 8],
                         [40, 60, 80]],
                        [[4, 5, 6],
                         [40, 50, 60]],
                        [9003,
                         8003]
                       ])

  def test_no_carry(self, backend):
    def _AddOne():  # pylint: disable=invalid-name
      return tl.Fn('AddOne', lambda x: x + 1)

    with fastmath.use_backend(backend):
      layer = tl.Scan(_AddOne(), n_carry=0)
      x = np.array([[1, 3, 7],
                    [10, 30, 70]])
      y = layer(x)
      self.assertEqual(as_list(y), [[2, 4, 8],
                                    [11, 31, 71]])


class CondTest(absltest.TestCase):

  def test_basic_true(self):
    cond = ReturnConst(True)
    true = ReturnConst([2])
    false = ReturnConst([5])
    layer = tl.Cond(cond, true, false)
    layer.init(())
    xs = tuple()
    ys = layer(xs)
    self.assertEqual(as_list(ys), 2)

  def test_basic_false(self):
    cond = ReturnConst(False)
    true = ReturnConst([2])
    false = ReturnConst([5])
    layer = tl.Cond(cond, true, false)
    layer.init(())
    xs = tuple()
    ys = layer(xs)
    self.assertEqual(as_list(ys), 5)

  def test_complex_blocks(self):
    cond = ReturnConst(True)
    true = DivideBy(2.)
    false = DivideBy(4.)
    layer = tl.Cond(cond, true, false)
    xs = [np.arange(5).astype(np.float32)]
    layer.init(shapes.signature(xs))
    ys = layer(xs)
    self.assertEqual(as_list(ys), [0., 0.5, 1.0, 1.5, 2.0])

  def test_condition_func_true(self):
    cond = SmallerThan(3.0)
    true = DivideBy(2.)
    false = DivideBy(4.)
    layer = tl.Cond(cond, true, false)
    xs = (np.array(2.), np.array([4., 12.]))
    layer.init(shapes.signature(xs))
    ys = layer(xs)
    self.assertEqual(as_list(ys), [2., 6.])

  def test_condition_func_false(self):
    cond = SmallerThan(3.0)
    true = DivideBy(2.)
    false = DivideBy(4.)
    layer = tl.Cond(cond, true, false)
    xs = (np.array(4.), np.array([4., 12.]))
    layer.init(shapes.signature(xs))
    ys = layer(xs)
    self.assertEqual(as_list(ys), [1., 3.])

  def test_condition_func_default_false(self):
    cond = SmallerThan(3.0)
    true = DivideBy(2.)
    layer = tl.Cond(cond, true)
    xs = (np.array(4.), np.array([4., 12.]))
    layer.init(shapes.signature(xs))
    ys = layer(xs)
    self.assertEqual(as_list(ys), [4., 12.])

  def test_exception_n_out(self):
    cond = SmallerThan(3.0)
    true = DivideBy(2.)
    false = tl.Dup()
    self.assertRaises(ValueError, lambda: tl.Cond(cond, true, false))

  def test_exception_n_in(self):
    cond = SmallerThan(3.0)
    true = ReturnConst(2.)
    false = DivideBy(2.)
    self.assertRaises(ValueError, lambda: tl.Cond(cond, true, false))

  def test_exception_run1(self):
    # We expect exactly one input.
    cond = SmallerThan(3.0)
    true = ReturnConst(2.)
    false = ReturnConst(5.)
    def init_and_run(layer, xs):
      layer.init(shapes.signature(xs))
      layer(xs)
    # It will pass with one input.
    xs = np.array(4.)
    layer = tl.Cond(cond, true, false)
    init_and_run(layer, xs)
    # It will fail with zero or two inputs.
    for xs in ((), (np.array(4.), np.array([4., 12.]))):
      layer = tl.Cond(cond, true, false)
      # pylint: disable=cell-var-from-loop
      self.assertRaises(Exception, lambda: init_and_run(layer, xs))

  def test_exception_run2(self):
    # We expect exactly two inputs.
    cond = SmallerThan(3.0)
    true = DivideBy(2.)
    false = DivideBy(5.)
    def init_and_run(layer, xs):
      layer.init(shapes.signature(xs))
      layer(xs)
    # It will pass with two inputs.
    xs = (np.array(4.), np.array([4., 12.]))
    layer = tl.Cond(cond, true, false)
    init_and_run(layer, xs)
    # It will fail with zero or one input.
    for xs in ((), (np.array(4.))):
      # pylint: disable=cell-var-from-loop
      self.assertRaises(Exception, lambda: init_and_run(layer, xs))

  def test_weights_and_state(self):
    cond = SmallerThan(3.0)
    true = tl.Dense(5)
    false = tl.Dense(5)
    different = tl.Dense(5)
    layer = tl.Cond(cond, true, false)
    xs = (np.array(2.), np.array([0., 1., 2.]))
    layer.init(shapes.signature(xs))

    # weights
    self.assertEqual(as_list(layer.weights),
                     as_list((cond.weights, true.weights, false.weights)))
    self.assertNotEqual(as_list(true.weights), as_list(false.weights))
    self.assertNotEqual(as_list(true.weights), as_list(different.weights))

    false.weights = true.weights
    self.assertEqual(as_list(layer.weights),
                     as_list((cond.weights, true.weights, true.weights)))

    layer.weights = (cond.weights, true.weights, different.weights)
    self.assertEqual(as_list(layer.weights),
                     as_list((cond.weights, true.weights, different.weights)))
    # state
    self.assertEqual(as_list(layer.state),
                     as_list((cond.state, true.state, false.state)))
    # just check if simple assignments (setter from base.Layer) work correctly
    # with Cond.init_weights_and_state ; all states are empty so there is no
    # point in checking equality
    false.state = true.state
    layer.state = (cond.state, true.state, different.state)


class BatchLeadingAxesTest(absltest.TestCase):

  def _Id3Dim(self):  # pylint: disable=invalid-name
    del self
    def f(x):
      assert len(x.shape) == 3
      return x
    return tl.Fn('Id3Dim', f, n_out=1)

  def test_2axes(self):
    layer = tl.BatchLeadingAxes(self._Id3Dim(), n_last_axes_to_keep=2)
    ys = layer(np.zeros((3, 4, 5)))
    self.assertEqual(ys.shape, (3, 4, 5))
    ys = layer(np.zeros((2, 3, 4, 5)))
    self.assertEqual(ys.shape, (2, 3, 4, 5))
    ys = layer(np.zeros((1, 2, 3, 4, 5)))
    self.assertEqual(ys.shape, (1, 2, 3, 4, 5))


class BidirectionalTest(absltest.TestCase):

  def test_dimensionality(self):
    x = np.ones((2, 3, 8))
    layer = tl.Bidirectional(tl.GRU(n_units=8))
    input_signature = shapes.signature(x)
    _, _ = layer.init(input_signature)
    yhat = layer(x)

    self.assertEqual(yhat.shape, (2, 3, 8 + 8))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.layers.test_utils."""

import functools

from absl.testing import absltest
import numpy as np

from trax.layers import test_utils
from trax.supervised import decoding


def arithmetic_sequence(input_seq, limit=10):
  # Increment the last symbol. Wrap to [0, 10).
  return (input_seq[-1] + 1) % limit


class TestUtilsTest(absltest.TestCase):

  def test_mock_transformer_lm_eval_equals_predict(self):
    model_fn = functools.partial(
        test_utils.MockTransformerLM,
        sequence_fn=arithmetic_sequence,
        vocab_size=10,
    )
    test_utils.test_eval_equals_predict_discrete(model_fn, vocab_size=10)

  def test_mock_transformer_lm_decodes_arithmetic_sequence(self):
    model = test_utils.MockTransformerLM(
        sequence_fn=arithmetic_sequence,
        vocab_size=10,
        mode='predict',
    )
    output = decoding.autoregressive_sample(
        model, max_length=5, start_id=0, eos_id=-1, accelerate=False
    )

    # Sequence including the leading 0 and the last predicted symbol.
    full_seq = list(range(6))
    # decoding.autoregressive_sample doesn't return the leading 0.
    np.testing.assert_array_equal(output, [full_seq[1:]])
    # The prediction buffers don't include the last predicted symbol.
    model.assert_prediction_buffers_equal([full_seq[:-1]])

  def test_mock_transformer_lm_rewinds(self):
    model = test_utils.MockTransformerLM(
        sequence_fn=arithmetic_sequence,
        vocab_size=10,
        mode='predict',
    )
    sample_3 = functools.partial(
        decoding.autoregressive_sample,
        max_length=3,
        eos_id=-1,
        accelerate=False,
    )

    # Generate the 3 initial symbols.
    init_output = sample_3(model, start_id=0)
    np.testing.assert_array_equal(init_output, [[1, 2, 3]])
    state = model.state

    # Generate the next 3 symbols.
    next_output = sample_3(model, start_id=init_output[0, -1])
    np.testing.assert_array_equal(next_output, [[4, 5, 6]])

    # Rewind and generate the last 3 symbols again.
    model.state = state
    next_output = sample_3(model, start_id=init_output[0, -1])
    np.testing.assert_array_equal(next_output, [[4, 5, 6]])

    # Check the buffers.
    model.assert_prediction_buffers_equal([[0, 1, 2, 3, 4, 5]])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax normalization layers."""

from trax.fastmath import numpy as jnp
from trax.layers import base


class BatchNorm(base.Layer):
  """Layer that performs batch normalization.

  In training, batch normalization keeps smoothed cumulative statistics across
  batches of input data and modifies each new batch so that its components are
  normally distributed. In eval or inference, a `BatchNorm` instance uses its
  stored mean and variance to approximately normalize each new batch of data.

  See https://arxiv.org/abs/1502.03167 for original presentation and motivation
  of batch normalization).
  """

  def __init__(self, axis=(0, 1, 2), epsilon=1e-5, center=True, scale=True,
               momentum=0.999, mode='train'):
    super().__init__()
    self._axis = axis
    self._epsilon = epsilon
    self._center = center
    self._scale = scale
    self._momentum = momentum
    self._mode = mode

  def forward(self, x):
    """Computes batch normalization as part of a forward pass in the model."""
    running_mean, running_var, n_batches = self.state
    if self._mode == 'train':
      n_batches += 1
      mean, var = self._fast_mean_and_variance(x)
      # Gather smoothed input statistics for later use in evals or inference.
      running_mean = _exponentially_smoothed(self._momentum, running_mean, mean)
      running_var = _exponentially_smoothed(self._momentum, running_var, var)
      self.state = (running_mean, running_var, n_batches)
    else:
      mean = running_mean
      var = running_var

    z = self._z_score(x, mean, var)
    beta, gamma = self._beta_gamma_with_correct_axes(x, self.weights)

    # Return the z rescaled by the parameters if requested.
    if self._center and self._scale:
      output = gamma * z + beta
    elif self._center:
      output = z + beta
    elif self._scale:
      output = gamma * z
    else:
      output = z
    if output.dtype != x.dtype:
      raise TypeError(f'The dtype of the output ({output.dtype}) of batch '
                      f'norm is not the same as the input ({x.dtype}). '
                      f'Batch norm should not change the dtype.')
    return output

  def init_weights_and_state(self, input_signature):
    """Helper to initialize batch norm weights and state."""
    axis = self._axis
    axis = (axis,) if jnp.isscalar(axis) else axis
    input_shape = input_signature.shape
    shape = tuple(d for i, d in enumerate(input_shape) if i not in axis)
    # TODO(jonni): Should beta and gamma match the dtype in the input signature?
    beta = jnp.zeros(shape, dtype='float32') if self._center else ()
    gamma = jnp.ones(shape, dtype='float32') if self._scale else ()
    def get_stats_axis(i, d):
      if i in axis:
        return 1
      else:
        return d
    stats_shape = tuple(get_stats_axis(i, d) for i, d in enumerate(input_shape))
    running_mean = jnp.zeros(stats_shape, dtype=jnp.float32)
    running_var = jnp.ones(stats_shape, dtype=jnp.float32)
    n_batches = jnp.zeros((), dtype=jnp.int64)
    self.weights = (beta, gamma)
    self.state = (running_mean, running_var, n_batches)

  def _fast_mean_and_variance(self, x):
    mean = jnp.mean(x, self._axis, keepdims=True)
    # Fast but less numerically-stable variance calculation than jnp.var.
    m1 = jnp.mean(x**2, self._axis, keepdims=True)
    variance = m1 - mean**2
    return mean, variance

  def _z_score(self, x, mean, variance):
    mu = mean.astype(x.dtype)
    sigma = jnp.sqrt(variance + self._epsilon).astype(x.dtype)
    return (x - mu) / sigma

  def _beta_gamma_with_correct_axes(self, x, weights):
    # Expand the parameters to have the right axes.
    beta, gamma = weights
    # TODO(phawkins): jnp.expand_dims should accept an axis tuple.
    # (https://github.com/numpy/numpy/issues/12290)
    ed = tuple(None if i in self._axis else slice(None)
               for i in range(jnp.ndim(x)))
    beta = beta[ed]
    gamma = gamma[ed]
    return beta, gamma


class LayerNorm(base.Layer):
  """Layer normalization."""

  def __init__(self, center=True, epsilon=1e-6):
    super().__init__()
    self._epsilon = epsilon
    self._center = center

  def forward(self, x):
    scale, bias = self.weights
    mean = jnp.mean(x, axis=-1, keepdims=True)
    centered = x - mean if self._center else x
    variance = jnp.mean(centered * centered, axis=-1, keepdims=True)
    norm_inputs = centered / jnp.sqrt(variance + self._epsilon)
    scaled = norm_inputs * scale
    return scaled + bias if self._center else scaled

  def init_weights_and_state(self, input_signature):
    features = input_signature.shape[-1]
    scale = jnp.ones(features, dtype=input_signature.dtype)
    bias = jnp.zeros(features, dtype=input_signature.dtype)
    self.weights = scale, bias


class FilterResponseNorm(base.Layer):
  """Filter Response Normalization layer without Threshold Linear Unit.

  c.f. https://arxiv.org/pdf/1911.09737.pdf
  """

  def __init__(self,
               mode=None,
               learn_epsilon=False,
               init_epsilon=1e-6,
               init_learnt_epsilon=1e-4):
    super().__init__()

    del mode

    # If we learn epsilon then epsilon = init_epsilon + |learnt_value|
    # where learnt_value is initialized to init_learnt_epsilon.
    # If learn_epsilon is false then epsilon is just init_epsilon.
    #
    # NOTE: I (afrozm) haven't been able to train with `learn_epsilon = True`.
    self._learn_epsilon = learn_epsilon

    # TODO(jonni): Replace asserts with ValueError.
    assert init_epsilon > 0
    assert init_learnt_epsilon > 0

    self._init_epsilon = jnp.array(init_epsilon, dtype=jnp.float32)
    self._init_learnt_epsilon = jnp.array(init_learnt_epsilon,
                                          dtype=jnp.float32)

  def forward(self, inputs):
    gamma, beta, epsilon_l = self.weights

    epsilon = self._init_epsilon
    if epsilon_l is not base.EMPTY_WEIGHTS:
      epsilon += jnp.abs(epsilon_l[0])

    # Omit B and C
    axis = tuple(range(1, len(jnp.shape(inputs)) - 1))
    # (B, 1, 1, C)
    nu2 = jnp.mean(inputs**2, axis=axis, keepdims=True)
    # (B, W, H, C)
    xhat = inputs / jnp.sqrt(nu2 + epsilon)

    return gamma * xhat + beta

  def init_weights_and_state(self, input_signature):
    # Usually (B, W, H, C)
    shape = input_signature.shape
    num_channels = shape[-1]

    gamma = jnp.ones((num_channels,), dtype=jnp.float32)
    beta = jnp.zeros((num_channels,), dtype=jnp.float32)

    epsilon_l = base.EMPTY_WEIGHTS
    if self._learn_epsilon:
      epsilon_l = (self._init_learnt_epsilon,)

    self.weights = gamma, beta, epsilon_l


def _exponentially_smoothed(momentum, old, new):
  smoothed_value = momentum * old + (1 - momentum) * new
  return smoothed_value.astype(old.dtype)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Adafactor optimizer class."""

from trax.fastmath import numpy as jnp
from trax.optimizers import base as opt_base


class Adafactor(opt_base.Optimizer):
  """Adafactor optimizer, as described in https://arxiv.org/abs/1804.04235."""

  def __init__(self,
               learning_rate=0.05,
               factored=True,
               multiply_by_parameter_scale=True,
               do_clipping=True,
               do_momentum=False,
               momentum_in_bfloat16=False,
               beta1=0.0,
               decay_rate=0.8,
               clipping_threshold=1.0,
               weight_decay_rate=1e-5,
               weight_decay_n_steps=0,
               epsilon1=1e-16,
               epsilon2=1e-3):
    """Create the Adafactor optimizer.

    Adafactor is described in https://arxiv.org/abs/1804.04235.

    Args:
      learning_rate: float: trax-provided learning rate.
      factored: boolean: whether to use factored second-moment estimator for 2d
        variables.
      multiply_by_parameter_scale: boolean: if True, then scale provided
        learning_rate by parameter norm. if False, provided learning_rate is
        absolute step size.
      do_clipping: whether to clip gradients; if True, set clipping_theshold.
      do_momentum: whether to use momentum; if True, set beta1.
      momentum_in_bfloat16: if True, store momentum in bfloat16 to save memory.
      beta1: a float value between 0 and 1, enables momentum and uses extra
        memory if nonzero!  Off by default.
      decay_rate: float: controls second-moment exponential decay schedule.
      clipping_threshold: an optional float >= 1, if None no update clipping.
      weight_decay_rate: rate at which to decay weights.
      weight_decay_n_steps: for how many steps to decay weights (always if None)
      epsilon1: Regularization constant for squared gradient.
      epsilon2: Regularization constant for parameter scale.
    """
    # These 4 parameters are not configurable once the class is created.
    self._factored = factored
    self._multiply_by_parameter_scale = multiply_by_parameter_scale
    self._do_clipping = do_clipping
    self._do_momentum = do_momentum
    self._momentum_in_bfloat16 = momentum_in_bfloat16
    # Dynamically configurable parameters will be passed to the update function.
    super().__init__(
        learning_rate=learning_rate,
        beta1=beta1,
        decay_rate=decay_rate,
        clipping_threshold=clipping_threshold,
        weight_decay_rate=weight_decay_rate,
        weight_decay_n_steps=weight_decay_n_steps,
        epsilon1=epsilon1,
        epsilon2=epsilon2,
    )

  @staticmethod
  def _decay_rate_pow(i, exponent=0.8):
    """Default Adafactor second-moment decay schedule."""
    t = jnp.array(i, jnp.float32) + 1.0
    return 1.0 - t**(-exponent)

  def init(self, weights):
    shape = weights.shape
    slots = []
    if self._factored and len(shape) >= 2:
      v_row = jnp.zeros(shape[:-1], dtype=jnp.float32)
      v_col = jnp.zeros(shape[:-2] + shape[-1:], dtype=jnp.float32)
      slots.extend([v_row, v_col])
    else:
      v = jnp.zeros_like(weights)
      slots.append(v)
    if self._do_momentum:
      m = jnp.zeros_like(weights)
      if self._momentum_in_bfloat16:
        m = m.astype(jnp.bfloat16)
      slots.append(m)
    return slots

  def update(self, step, grads, weights, slots, opt_params):
    updates = []
    learning_rate = opt_params['learning_rate']
    beta1 = opt_params['beta1']
    decay_rate = opt_params['decay_rate']
    clipping_threshold = opt_params['clipping_threshold']
    weight_decay_rate = opt_params['weight_decay_rate']
    weight_decay_n_steps = opt_params['weight_decay_n_steps']
    weight_decay_rate = jnp.where(
        weight_decay_n_steps < 1,  # if weight_decay_n_steps == 0, ignore it
        weight_decay_rate,
        (weight_decay_rate * jnp.maximum(weight_decay_n_steps - step, 0.0) /
         jnp.maximum(weight_decay_n_steps, 0.0)))
    epsilon1 = opt_params['epsilon1']
    epsilon2 = opt_params['epsilon2']
    decay_rate = self._decay_rate_pow(step, exponent=decay_rate)
    update_scale = learning_rate
    if self._multiply_by_parameter_scale:
      update_scale *= jnp.maximum(
          jnp.sqrt(jnp.mean(weights * weights)), epsilon2)
    mixing_rate = 1.0 - decay_rate

    grads_sqr = grads * grads
    if self._factored and len(weights.shape) >= 2:
      v_row = slots[0]  # In this case, the slots are (v_row, v_col, ...).
      v_col = slots[1]
      new_v_row = (
          decay_rate * v_row + mixing_rate * jnp.mean(grads_sqr, axis=-1))
      new_v_col = (
          decay_rate * v_col + mixing_rate * jnp.mean(grads_sqr, axis=-2))
      updates.extend([new_v_row, new_v_col])
      row_mean = jnp.mean(new_v_row, axis=-1, keepdims=True)
      row_factor = (row_mean / (new_v_row + epsilon1))**0.5
      col_factor = (new_v_col + epsilon1)**-0.5
      y = (
          grads * jnp.expand_dims(row_factor, axis=-1) *
          jnp.expand_dims(col_factor, axis=-2))
    else:
      v = slots[0]  # In this case, the slots are (v, ...)
      new_v = decay_rate * v + mixing_rate * grads_sqr
      updates.append(new_v)
      y = grads * (new_v + epsilon1)**-0.5

    if self._do_clipping:
      clipping_denom = (
          jnp.maximum(1.0, jnp.sqrt(jnp.mean(y * y)) / clipping_threshold))
      y /= clipping_denom

    subtrahend = update_scale * y
    if self._do_momentum:
      m = slots[-1]  # Momentum is always the last slot (if used).
      m = m.astype(subtrahend.dtype)  # Accumulate in subtrahend dtype.
      new_m = beta1 * m + (1.0 - beta1) * subtrahend
      subtrahend = new_m
      updates.append(new_m.astype(slots[-1].dtype))  # Back to bfloat if needed.

    new_weights = (1 - weight_decay_rate) * weights - subtrahend
    # TODO(lukaszkaiser): why is the astype needed here? Check and correct.
    return new_weights.astype(weights.dtype), updates



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for Reformer models."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
import gin
import numpy as np

from trax import fastmath
from trax import layers as tl
from trax import shapes
from trax.models.reformer import reformer


BACKENDS = [fastmath.Backend.JAX]


def short_name(b):
  if b == fastmath.Backend.JAX:
    return 'jax'
  else:
    return 'tf'


class ReformerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    gin.clear_config()

  def _lsh_self_attention_fn(self):
    return functools.partial(
        tl.LSHSelfAttention,
        attention_dropout=0.0,
        chunk_len=64,
        n_buckets=[32, 32],
        n_chunks_after=0,
        n_chunks_before=1,
        n_hashes=1,
        n_parallel_heads=1,
        predict_drop_len=128,
        predict_mem_len=1024,
    )

  def _timebin_self_attention_fn(self, use_reference_code=False):
    return functools.partial(
        tl.SelfAttention,
        attention_dropout=0.05,
        chunk_len=64,
        n_chunks_before=1,
        n_parallel_heads=1,
        use_reference_code=use_reference_code
    )

  def test_reformer_lm_forward_shape(self):
    vocab_size = 16
    model = reformer.ReformerLM(
        vocab_size, d_model=32, d_ff=64, d_attention_key=16,
        d_attention_value=16, n_layers=1, n_heads=2, max_len=16)
    xs = [np.ones((1, 8)).astype(np.int32),
          np.ones((1, 8)).astype(np.int32)]
    _, _ = model.init(shapes.signature(xs))
    ys = model(xs)
    self.assertEqual([y.shape for y in ys], [(1, 8, 16), (1, 8)])


  def test_reformer_lm_lsh(self):
    lsh_self_attention = self._lsh_self_attention_fn()
    timebin_self_attention = self._timebin_self_attention_fn()

    model = reformer.ReformerLM(
        vocab_size=256,
        d_model=256,
        d_ff=512,
        d_attention_key=64,
        d_attention_value=64,
        n_layers=2,
        n_heads=2,
        dropout=0.05,
        max_len=65536,
        attention_type=[timebin_self_attention, lsh_self_attention],
        pos_axial_shape=(256, 256),
        pos_d_axial_embs=(64, 192),
        ff_activation=tl.Relu,
        ff_use_sru=0,
        ff_chunk_size=8192,
        mode='train',
    )
    x = np.ones((1, 65536)).astype(np.int32)
    weights, state = model.init(shapes.signature(x))

    @fastmath.jit
    def mock_training_step(x, weights, state, rng):
      def compute_mock_loss(weights):
        logits, new_state = model.pure_fn(x, weights, state, rng)
        loss = fastmath.numpy.mean(logits[..., 0])
        return loss, (new_state, logits)
      gradients, (new_state, logits) = fastmath.grad(
          compute_mock_loss, has_aux=True)(weights)
      new_weights = fastmath.nested_map_multiarg(
          lambda w, g: w - 1e-4 * g, weights, gradients)
      return new_weights, new_state, logits

    weights, state, logits = mock_training_step(
        x, weights, state, fastmath.random.get_prng(0))
    self.assertEqual(logits.shape, (1, 65536, 256))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for conv layers."""

from absl.testing import absltest
import numpy as np

import trax.layers as tl


class MaxPoolTest(absltest.TestCase):

  def test_forward_shape(self):
    layer = tl.MaxPool(pool_size=(2, 2), strides=(1, 2))
    x = np.ones((11, 6, 4, 17))
    y = layer(x)
    self.assertEqual(y.shape, (11, 5, 2, 17))

  def test_forward(self):
    layer = tl.MaxPool(pool_size=(2, 2), strides=(2, 2))
    x = np.array([[
        [[1, 2, 3], [4, 5, 6], [10, 20, 30], [40, 50, 60]],
        [[4, 2, 3], [7, 1, 2], [40, 20, 30], [70, 10, 20]],
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[[7, 5, 6], [70, 50, 60]]]])

  def test_padding_default(self):
    layer = tl.MaxPool(pool_size=(3,), strides=(3,))

    # Discard incomplete window at end: [[3, 6], [4, 5]].
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6], [4, 5]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[2, 9]]])

  def test_padding_same(self):
    layer = tl.MaxPool(pool_size=(3,), strides=(3,), padding='SAME')

    # One padding position needed; add at end.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6], [4, 5]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[2, 9], [4, 6]]])

    # Two padding positions needed; add one at end and one at start.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[1, 9], [3, 7]]])


class SumPoolTest(absltest.TestCase):

  def test_forward_shape(self):
    layer = tl.SumPool(pool_size=(2, 2), strides=(1, 2))
    x = np.ones((11, 6, 4, 17))
    y = layer(x)
    self.assertEqual(y.shape, (11, 5, 2, 17))

  def test_forward(self):
    layer = tl.SumPool(pool_size=(2, 2), strides=(2, 2))
    x = np.array([[
        [[1, 2, 3], [4, 5, 6], [10, 20, 30], [40, 50, 60]],
        [[4, 2, 3], [7, 1, 2], [40, 20, 30], [70, 10, 20]],
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[[16, 10, 14], [160, 100, 140]]]])

  def test_padding_same(self):
    layer = tl.SumPool(pool_size=(3,), strides=(3,), padding='SAME')

    # One padding position needed; add at end.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6], [4, 5]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[3, 24], [7, 11]]])

    # Two padding positions needed; add one at end and one at start.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[1, 17], [5, 13]]])


class AvgPoolTest(absltest.TestCase):

  def test_forward_shape(self):
    layer = tl.AvgPool(pool_size=(2, 2), strides=(1, 2))
    x = np.ones((11, 6, 4, 17))
    y = layer(x)
    self.assertEqual(y.shape, (11, 5, 2, 17))

  def test_forward(self):
    layer = tl.AvgPool(pool_size=(2, 2), strides=(2, 2))
    x = np.array([[
        [[1, 2, 3], [4, 5, 6], [10, 20, 30], [40, 50, 60]],
        [[4, 2, 3], [7, 1, 2], [40, 20, 30], [70, 10, 20]],
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[[4.0, 2.5, 3.5], [40, 25, 35]]]])

  def test_padding_same(self):
    layer = tl.AvgPool(pool_size=(3,), strides=(3,), padding='SAME')

    # One padding position needed; add at end.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6], [4, 5]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[1, 8], [3.5, 5.5]]])

    # Two padding positions needed; add one at end and one at start.
    x = np.array([[
        [0, 9], [1, 8], [2, 7], [3, 6]
    ]])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [[[.5, 8.5], [2.5, 6.5]]])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.supervised.callbacks."""

import functools
import io
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import gym
import numpy as np

from trax import models
from trax import test_utils
from trax.data import inputs
from trax.layers import test_utils as tl_test_utils
from trax.rl import serialization_utils
from trax.rl import space_serializer
from trax.supervised import callbacks
from trax.supervised import lr_schedules
from trax.supervised import trainer_lib
from trax.supervised import training


def random_inputs(seq_len, batch_size):
  def stream_fn(num_devices):
    del num_devices
    while True:
      x = np.random.uniform(size=(batch_size, seq_len))
      y = np.random.uniform(size=(batch_size, seq_len))
      mask = np.ones_like(x).astype(np.float32)
      yield (x, y, x, mask)

  return inputs.Inputs(
      train_stream=stream_fn,
      eval_stream=stream_fn,
  )


def make_multibonacci_modulo(history_length, limit):
  """Creates a function that generates the Multibonacci sequence modulo n."""
  def sequence_fn(seq):
    return np.sum(seq[-history_length:]) % limit
  return sequence_fn


def generate_trajectory(sequence_fn, space, n_steps):
  """Generates random actions and observations that follow sequence_fn."""
  act = [space.sample() for _ in range(n_steps)]
  obs = [space.sample()]

  for (o, a) in zip(
      obs,
      act[:-1],  # Don't generate the last observation.
  ):
    context = list(np.array([o, a]).flatten())
    symbols = []
    for _ in range(np.array(o).size):
      symbol = sequence_fn(context + symbols)
      symbols.append(symbol)
    obs.append(np.reshape(symbols, space.shape))

  obs = np.array([obs])
  act = np.array([act])
  return (obs, act)


def make_singleton_eval_task(observations, actions):
  """Creates an EvalTask with just one example."""
  mask = np.ones(observations.shape[:2])
  def data():
    while True:
      yield (observations, actions, observations, mask)

  return training.EvalTask(
      labeled_data=data(),
      metrics=[],
  )


def make_serialized_model(seq_model, space, vocab_size):
  srl = space_serializer.create(space, vocab_size)
  return serialization_utils.SerializedModel(
      functools.partial(seq_model, vocab_size=vocab_size),
      observation_serializer=srl,
      action_serializer=srl,
      significance_decay=0.7,
  )


class CallbacksTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')

  @mock.patch('sys.stdout', new_callable=io.StringIO)
  def test_serialized_model_evaluation(self, mock_stdout):
    precision = 1
    vocab_size = 2
    srl = space_serializer.BoxSpaceSerializer(
        space=gym.spaces.Box(shape=(), low=0.0, high=1.0),
        vocab_size=vocab_size,
        precision=precision,
    )

    def inner_model(mode):
      return models.TransformerLM(
          mode=mode,
          vocab_size=vocab_size,
          d_model=2,
          d_ff=4,
          n_layers=1,
          n_heads=1,
      )

    serialized_model_fn = functools.partial(
        serialization_utils.SerializedModel,
        inner_model,
        observation_serializer=srl,
        action_serializer=srl,
        significance_decay=0.7,
    )
    eval_callback = functools.partial(
        callbacks.SerializedModelEvaluation, eval_at=5
    )

    output_dir = self.create_tempdir().full_path
    trainer_lib.train(
        output_dir=output_dir,
        model=serialized_model_fn,
        inputs=functools.partial(random_inputs, seq_len=4, batch_size=64),
        lr_schedule_fn=functools.partial(lr_schedules.constant, 0.01),
        callbacks=[eval_callback],
        steps=10,
    )
    self.assertTrue(_has_metric('pred_error', mock_stdout))

  @parameterized.product(
      context_lengths=((2,), (1, 3)),
      horizon_lengths=((1,), (1, 2)),
  )
  def test_srl_eval_feeds_correct_sequence(
      self, context_lengths, horizon_lengths
  ):
    vocab_size = 10
    n_steps = 5

    multibonacci_modulo = make_multibonacci_modulo(2, vocab_size)
    space = gym.spaces.Discrete(n=vocab_size)
    (obs, act) = generate_trajectory(multibonacci_modulo, space, n_steps)
    eval_task = make_singleton_eval_task(obs, act)
    seq_model = functools.partial(
        tl_test_utils.MockTransformerLM,
        sequence_fn=multibonacci_modulo,
    )
    serialized_model = make_serialized_model(seq_model, space, vocab_size)
    callback = callbacks.SerializedModelEvaluation(
        loop=None,
        eval_task=eval_task,
        model=serialized_model,
        context_lengths=context_lengths,
        horizon_lengths=horizon_lengths,
        accelerate_model=False,
    )
    callback.evaluate(weights=None)

    expected_seq = np.zeros(2 * n_steps + 1)
    expected_seq[1::2] = obs
    expected_seq[2::2] = act
    seen_len = (context_lengths[-1] + horizon_lengths[-1]) * 2
    callback.predict_model.assert_prediction_buffers_equal(
        [expected_seq[:seen_len]]
    )

  @parameterized.named_parameters(('one_symbol', 1), ('two_symbols', 2))
  def test_srl_eval_reports_zero_error_for_perfect_model(self, precision):
    vocab_size = 100
    n_steps = 5

    multibonacci_modulo = make_multibonacci_modulo(2 * precision, vocab_size)
    space = gym.spaces.MultiDiscrete(nvec=([vocab_size] * precision))
    (obs, act) = generate_trajectory(multibonacci_modulo, space, n_steps)
    eval_task = make_singleton_eval_task(obs, act)
    seq_model = functools.partial(
        tl_test_utils.MockTransformerLM,
        sequence_fn=multibonacci_modulo,
    )
    serialized_model = make_serialized_model(seq_model, space, vocab_size)
    callback = callbacks.SerializedModelEvaluation(
        loop=None,
        eval_task=eval_task,
        model=serialized_model,
        context_lengths=(1,),
        horizon_lengths=(4,),
        accelerate_model=False,
    )
    metrics = callback.evaluate(weights=None)
    error = next(
        value for (name, value) in metrics.items() if 'pred_error' in name
    )
    assert error == 0


def _has_metric(metric_name, stdout):
  log = stdout.getvalue()
  metric_logs = [line for line in log.split('\n') if metric_name in line]
  return bool(metric_logs)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Layers that can run in reverse to compute inputs from outputs.

Reversible layers reduce the memory required for backpropagation-based
training, especially for *deep* networks. In a series of reversible layers,
input activations from a forward pass don't need to be stored: they can be
reconstructed on the backward pass, layer by layer, from outputs to inputs.

See, e.g., [The Reversible Residual Network: Backpropagation Without Storing
Activations](https://arxiv.org/abs/1707.04585) and [Reformer: The Efficient
Transformer](https://arxiv.org/abs/2001.04451).
"""

from absl import logging
import jax
from trax import fastmath
from trax.layers import base
from trax.layers import combinators as cb


_split_rngs = cb._split_rngs  # pylint: disable=protected-access


class ReversibleLayer(base.Layer):
  """Reversible Layer."""

  def reverse(self, output, weights=(), state=(), new_state=(), rng=None):
    """Reverse this layer: compute input given output."""
    raise NotImplementedError

  def _pure_forward(self, x, weights, state, rng):
    """Call self.forward in a pure way."""
    old_weights, old_state, old_rng = self.weights, self.state, self._rng
    self.weights, self.state, self._rng = weights, state, rng
    res = self.forward(x)
    self.weights, self.state, self._rng = old_weights, old_state, old_rng
    return res

  def reverse_and_grad(self, output, grad, weights=(), state=(), new_state=(),
                       rng=None):
    """Backward pass: computes the inverse of a layer and propagates gradients.

    While you may choose to only implement reverse, some layers implement this
    function directly as computation may be shared between reversing and
    computing gradients.

    Args:
      output: Output activations; can be a (possibly nested) tuple.
      grad: gradient signal (cotangent) computed based on subsequent layers.
        The structure and shape must match the output.
      weights: layer weights
      state: start state
      new_state: updated state computed by the forward pass
      rng: Single-use random number generator (JAX PRNG key).

    Returns:
      A tuple (x, (x_grad, weights_grad)), where x is the reconstructed input,
      x_grad is the gradient signal for the input, and weights_grad is the
      gradient signal for the weights.
    """
    reconstructed_x = self.reverse(output, weights, state, new_state, rng)
    _, vjpfun = fastmath.vjp(
        self._pure_forward, reconstructed_x, weights, state, rng)
    x_grad, weights_grad, _, _ = vjpfun(grad)
    return reconstructed_x, (x_grad, weights_grad)

  @property
  def has_backward(self):
    return True

  def backward(self, inputs, output, grad, weights, state, new_state, rng):
    del inputs
    _, inputs_weights_grad = (
        self.reverse_and_grad(output, grad, weights, state, new_state, rng))
    return inputs_weights_grad


class ReversibleConcatenatePair(ReversibleLayer):
  """Maps (x, y) -> ([x, y], [x, y]);  [x, y] is concatenation on last axis."""

  def __init__(self):
    super().__init__(n_in=2, n_out=2)

  def forward(self, inputs):
    x, y = inputs
    r = fastmath.numpy.concatenate((x, y), axis=-1)
    return r, r

  def reverse(self, outputs, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    pair, _ = outputs
    x, y = fastmath.numpy.split(pair, 2, axis=-1)
    return x, y


class ReversibleSelect(ReversibleLayer):
  """Reversible version of the Select combinator."""

  def __init__(self, indices, n_in=None, name=None):
    if n_in is None:
      n_in = max(indices) + 1
    if name is None:
      name = f'ReversibleSelect{indices}'.replace(' ', '')
    super().__init__(n_in=n_in, n_out=len(indices), name=name)
    self._indices = indices

    # Calculate reverse indices.
    self._reverse_indices = []
    for i in range(n_in):
      if i not in indices:
        raise ValueError('To be reversible, all inputs to Select must be in '
                         'indices. Did not find %d in indices.' % i)
      else:
        self._reverse_indices.append(indices.index(i))

  def forward(self, inputs):
    if not isinstance(inputs, (tuple, list)):
      inputs = (inputs,)
    selected = tuple(inputs[i] for i in self._indices)
    return selected[0] if len(selected) == 1 else selected

  def reverse(self, outputs, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    if not isinstance(outputs, (tuple, list)):
      outputs = (outputs,)
    selected = tuple(outputs[i] for i in self._reverse_indices)
    return selected[0] if len(selected) == 1 else selected


def ReversibleSwap():  # pylint: disable=invalid-name
  return ReversibleSelect([1, 0], name='ReversibleSwap')


class ReversibleReshape(ReversibleLayer):
  """Reversible reshaping layer."""

  def __init__(self, shape1, shape2, n_in=1):
    self._shape1 = list(shape1)
    self._shape2 = list(shape2)
    name = 'ReversibleReshape_%s_%s' % (str(shape1), str(shape2))
    super().__init__(n_in=n_in, n_out=n_in, name=name)

  def forward(self, inputs):
    if not isinstance(inputs, (tuple, list)):
      inputs = (inputs,)
    res = []
    for x in inputs:
      new_shape = self._shape2 + list(x.shape)[len(self._shape1):]
      res.append(fastmath.numpy.reshape(x, new_shape))
    if len(res) == 1:
      return res[0]
    return tuple(res)

  def reverse(self, outputs, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    if not isinstance(outputs, (tuple, list)):
      outputs = (outputs,)
    res = []
    for x in outputs:
      new_shape = self._shape1 + list(x.shape)[len(self._shape2):]
      res.append(fastmath.numpy.reshape(x, new_shape))
    if len(res) == 1:
      return res[0]
    return tuple(res)


class ReversiblePrintShape(ReversibleLayer):
  """Reversible PrintShape for debugging reversible serial layers."""

  def __init__(self, n_in=1, msg=''):
    super().__init__(n_in=n_in, n_out=n_in)
    self._msg = msg

  def forward(self, xs):
    shapes_and_dtypes = ', '.join([str(x.shape) + f'[{x.dtype}]' for x in xs])
    info = f'PrintShape: {self._msg}: [{shapes_and_dtypes}]'
    print(info)
    logging.info(info)
    return xs

  def reverse(self, outputs, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    return outputs


class ReversibleSerial(ReversibleLayer, cb.Serial):
  """A reversible version of tl.Serial (requires reversible sub-layers)."""

  def __init__(self, *layers):
    super().__init__(*layers)
  # def __init__(self, *layers):  # pylint: disable=super-init-not-called
  #   cb.Serial.__init__(self, layers)

    # Note that sublayers has already been flattened to remove nested lists.
    for i, layer in enumerate(self.sublayers):
      if not isinstance(layer, ReversibleLayer):
        raise ValueError(
            'Sub-layer {} of ReversibleSerial is not reversible: {}'.format(
                i, layer))

  def reverse(self, output, weights=(), state=(), new_state=(), rng=None):
    rngs = (None,) * self._n_layers
    if rng is not None:
      rngs = fastmath.random.split(rng, self._n_layers)

    stack = output
    for layer, p, s, ns, rng in reversed(list(zip(
        self.sublayers, weights, state, new_state, rngs))):
      layer_val = cb.inputs_from_stack(stack, layer.n_out)
      layer_val = layer.reverse(layer_val, p, s, ns, rng=rng)
      stack = cb.outputs_onto_stack(layer_val, stack, layer.n_out)

    return stack

  def reverse_and_grad(self, output, grad, weights=(), state=(), new_state=(),
                       rng=None):
    rngs = (None,) * self._n_layers
    if rng is not None:
      rngs = fastmath.random.split(rng, self._n_layers)

    stack = output
    stack_grad = grad
    weights_grad = []
    for layer, p, s, ns, rng in reversed(list(zip(
        self.sublayers, weights, state, new_state, rngs))):
      layer_val = cb.inputs_from_stack(stack, layer.n_out)
      layer_ct = cb.inputs_from_stack(stack_grad, layer.n_out)
      layer_val, layer_ct = layer.reverse_and_grad(
          layer_val, layer_ct, p, s, ns, rng=rng)
      layer_ct, p_ct = layer_ct
      weights_grad.insert(0, p_ct)
      stack = cb.outputs_onto_stack(layer_val, stack, layer.n_out)
      stack_grad = cb.outputs_onto_stack(layer_ct, stack_grad, layer.n_out)

    return stack, (stack_grad, tuple(weights_grad))


class ReversibleHalfResidual(ReversibleLayer):
  """Half of a RevNet-style residual that optionally performs attention.

  When attention_layer is None, this layer has the signature ::

      [accumulator, *context] -> [accumulator + f(context), *context]

  The attention_layer must be an instance of EfficientAttentionBase or one of
  its subclasses (see efficient_attention.py), or None.

  Attention is special-cased for the following two reasons:

  - LSH attention needs to save bucket assignments from the forward pass to the
    backward pass, for training stability. This requires special-casing it.
  - We can call attention_layer.forward_and_or_backward to compute its output
    (needed for inverting a reversible residual layer) while simultaneously
    performing the backward pass. Sharing computation between these two
    operations improves training speed.
  """

  def __init__(self, *residual_layers, attention_layer=None, name=None):
    super().__init__(name=name)

    self._compute_residual = cb.Serial(*residual_layers)
    self._attention_layer = attention_layer

    if self._attention_layer is None:
      self._sublayers = (self._compute_residual,)
    else:
      if hasattr(attention_layer, 'forward_and_or_backward'):
        self._forward_and_or_backward = attention_layer.forward_and_or_backward
      else:
        self._forward_and_or_backward = _forward_and_or_backward(
            attention_layer)
      self._sublayers = (self._compute_residual, self._attention_layer)

    running_max = 0
    running_total = 0
    for layer in self._sublayers:
      running_total += layer.n_in
      running_max = max(running_max, running_total)
      running_total -= layer.n_out
    self._n_in = self._n_out = running_max + 1

  def forward(self, xs):
    rngs = _split_rngs(self.rng, len(self.sublayers))
    accumulator, *context = xs
    stack = context = tuple(context)
    new_state = []
    for layer, w, s, rng in zip(self.sublayers, self.weights, self.state, rngs):
      inputs = cb.inputs_from_stack(stack, layer.n_in)
      if base.N_WEIGHTS_SHARDS > 1:
        # With sharded weights, make sure we don't keep them concatenated
        # in memory on each device by using remat.
        outputs, s = jax.remat(layer.pure_fn)(inputs, w, s, rng)
      else:
        outputs, s = layer.pure_fn(inputs, w, s, rng)
      stack = cb.outputs_onto_stack(outputs, stack, layer.n_in)
      new_state.append(s)
    residual = stack[0] if isinstance(stack, (tuple, list)) else stack

    output = accumulator + residual
    stack = (output,) + context
    self.state = tuple(new_state)
    return stack

  def reverse(self, output, weights=(), state=(), new_state=(), rng=None):
    raise NotImplementedError('Only reverse_and_grad is actually used.')

  def reverse_and_grad(self, output, ct, weights=(), state=(), new_state=(),
                       rng=None):
    rngs = _split_rngs(rng, len(self.sublayers))

    accumulator_output, *context = output
    context = tuple(context)
    accumulator_output_ct, *context_ct = ct
    context_ct = tuple(context_ct)

    # Forward pass through self._compute_residual. Outputs that will not receive
    # a gradient signal from subsequent layers are moved to aux.
    def call_compute_residual(x, weights):
      state_to_pass = state[0]  # old_state

      # _replace_second_time is currently used exclusively in _RememberInReverse
      # layer to combat numerical instability in Terraformer when quantizing
      # the mask in SparseFF.
      def _replace_second_time(stt, nstt):
        if (isinstance(stt, tuple) and len(stt) == 2 and
            isinstance(stt[1], dict) and 'running_second_time' in stt[1]):
          return (nstt[0], {'running_second_time_yes': ()})
        elif isinstance(stt, (tuple, list)):
          assert isinstance(nstt, (tuple, list)) and len(nstt) == len(stt)
          return type(stt)([
              _replace_second_time(s, ns) for s, ns in zip(stt, nstt)])
        else:
          return stt

      state_to_pass = _replace_second_time(state_to_pass, new_state[0])
      res, _ = self._compute_residual.pure_fn(
          x, weights=weights, state=state_to_pass, rng=rngs[0])
      if not isinstance(res, (tuple, list)):
        return res, None
      else:
        n_differentiable = 1
        if self._attention_layer is not None:
          n_differentiable = min(len(res), self._attention_layer.n_in)
        return res[:n_differentiable], res[n_differentiable:]

    stack = context
    inputs = cb.inputs_from_stack(stack, self._compute_residual.n_in)
    outputs, compute_residual_vjpfun, outputs_aux = fastmath.vjp(
        call_compute_residual, inputs, weights[0], has_aux=True)
    if outputs_aux is not None:
      n_differentiable_outputs = len(outputs)
      outputs = outputs + outputs_aux
    stack = cb.outputs_onto_stack(outputs, stack, self._compute_residual.n_in)

    stack_ct = accumulator_output_ct
    if self._attention_layer is None:
      residual = stack[0] if isinstance(stack, (tuple, list)) else stack
    else:
      inputs = cb.inputs_from_stack(stack, self._attention_layer.n_in)
      (residual, _, attn_inputs_ct, attn_weights_ct
      ) = self._forward_and_or_backward(
          inputs, weights[1], new_state[1], rngs[1],
          output_grad=accumulator_output_ct,
          compute_output=True, update_state=False)
      stack_ct = cb.outputs_onto_stack(
          attn_inputs_ct, stack_ct, self._attention_layer.n_out)

    compute_residual_ct = cb.inputs_from_stack(
        stack_ct, self._compute_residual.n_out)
    if outputs_aux is not None:
      if not isinstance(compute_residual_ct, (tuple, list)):
        compute_residual_ct = (compute_residual_ct,)
      compute_residual_ct = compute_residual_ct[:n_differentiable_outputs]
      assert len(compute_residual_ct) == n_differentiable_outputs
    (compute_residual_inputs_ct, compute_residual_weights_ct
    ) = compute_residual_vjpfun(compute_residual_ct)
    stack_ct = cb.outputs_onto_stack(
        compute_residual_inputs_ct, stack_ct, self._compute_residual.n_out)
    if not isinstance(stack_ct, (tuple, list)):
      stack_ct = (stack_ct,)
    def _add(x, y):
      # `None` is for TFNP backend, which uses `None` as the gradient of
      # int/bool instead of an array of dtype `float0`.
      if x is None or x.dtype == jax.float0:
        return y
      if y is None or y.dtype == jax.float0:
        return x
      return x + y
    stack_ct = (accumulator_output_ct,) + fastmath.nested_map_multiarg(
        _add, context_ct[:len(stack_ct)], stack_ct) + context_ct[len(stack_ct):]

    reconstructed_x = accumulator_output - residual
    stack = (reconstructed_x,) + context
    if self._attention_layer is None:
      weights_ct = (compute_residual_weights_ct,)
    else:
      weights_ct = (compute_residual_weights_ct, attn_weights_ct)
    return stack, (stack_ct, weights_ct)

  # pylint: disable=protected-access
  def init_weights_and_state(self, input_signature):
    stack = input_signature[1:]
    if len(stack) == 1:
      stack = stack[0]

    inputs = cb.inputs_from_stack(stack, self._compute_residual.n_in)
    weights, state = self._compute_residual.init(inputs)
    outputs, _ = self._compute_residual._forward_abstract(inputs)
    stack = cb.outputs_onto_stack(outputs, stack, self._compute_residual.n_in)

    if self._attention_layer is None:
      self.state = (state,)
      self.weights = (weights,)
    else:
      inputs = cb.inputs_from_stack(stack, self._attention_layer.n_in)
      attn_weights, attn_state = self._attention_layer.init(inputs)
      self.state = (state, attn_state)
      self.weights = (weights, attn_weights)
  # pylint: enable=protected-access


def _forward_and_or_backward(layer):
  """Create forward_and_or_backward for layers that don't define it."""

  def forward_and_or_backward(inputs, weights, state, rng, output_grad=None,
                              compute_output=True, update_state=True):
    """Performs batched forward and/or backward passes.

    Args:
      inputs: inputs to the attention layer
      weights: weights for the attention layer
      state: state of the attention layer
      rng: PRNG key for the layer (shared across all examples and heads)
      output_grad: gradient of the loss wrt the output of the layer, or None.
          This function performs the backward pass iff `output_grad` is not
          None.
      compute_output: bool: whether to return the output of the forward pass
          (for example, a pure backwards pass does not need to return the
          output).
      update_state: bool: whether to return an updated layer state.

    Returns:
      A tuple (output, new_state, inputs_grad, weights_grad).
      - output is not None iff compute_output is True
      - new_state is not None iff update_state is True
      - inputs_grad & weights_grad are not None iff output_grad is not None
    """
    # Calculate the vector-Jacobian product of the layer pure_fn.
    output, vjp_fn, new_state = fastmath.vjp(
        layer.pure_fn, inputs, weights, state, rng, has_aux=True)
    output = output if compute_output else None
    new_state = new_state if update_state else None

    # The vjp function returns gradients with respect to inputs and weights.
    if output_grad is not None:
      grads_inputs, grads_weights, _, _ = vjp_fn(output_grad)
    else:
      grads_inputs, grads_weights = None, None

    return (output, new_state, grads_inputs, grads_weights)
  return forward_and_or_backward



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Residual Shuffle-Exchange Networks.

https://arxiv.org/pdf/2004.04662.pdf
"""

import numpy as np
from trax import fastmath
from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.layers.assert_shape import assert_shape


# pylint: disable=invalid-name
def _inverse_sigmoid(x):
  return np.log(x / (1 - x))


@assert_shape('...->...')
class _ClippedScaling(tl.Layer):
  """Pointwise multiplies by sigmoid(S) with a learnable vector S."""

  def __init__(self,
               residual_weight):
    super().__init__(n_in=1, n_out=1)
    self._residual_weight = residual_weight

  def forward(self, x):
    s = self.weights
    return jnp.multiply(x, fastmath.expit(s))

  def init_weights_and_state(self, input_signature):
    self.weights = _inverse_sigmoid(self._residual_weight) * np.ones(
        (input_signature.shape[-1])).astype('float32')


@assert_shape('bld->bld')
def ResidualSwitchUnit(
    d_model, dropout=0.1, mode='train', residual_weight=0.9):
  r"""RSU (Residual Switch Unit) layer as in https://arxiv.org/pdf/2004.04662.pdf.

  As defined in the paper:

  .. math::
    i &= [i_1, i_2] \\
    g &= GELU(LayerNorm(Z i)) \\
    c &= W g + B \\
    [o_1, o_2] &= \sigma(S) \bigodot i + h \bigodot c

  where Z, W, B, S are learnable parameters with sizes 2m × 4m, 4m × 2m, 2m, 2m.
  We assume that both i_1 and i_2 have size m. h is a scalar value.

  We assume the input is of shape [batch, length, depth].

  Args:
    d_model: output depth of the SRU layer
    dropout: dropout rate used in 'train' mode
    mode: mode for dropout layer
    residual_weight: value used in initializing vector S and constant h

  Returns:
    The RSU layer.
  """
  return tl.Serial(
      tl.Fn(
          'Reshape2Pairs',
          lambda x: jnp.reshape(x, (x.shape[0], x.shape[1] // 2, -1)),
          n_out=1),
      tl.Residual(
          tl.Dense(4 * d_model, use_bias=False),
          tl.LayerNorm(),
          tl.Gelu(),
          tl.Dense(2 * d_model),
          tl.Fn('Scaling',
                lambda x: x * np.sqrt(1 - residual_weight**2) * 0.25,
                n_out=1),
          shortcut=_ClippedScaling(residual_weight)),
      tl.Fn(
          'UnPair',
          lambda x: jnp.reshape(x, (x.shape[0], x.shape[1] * 2, -1)),
          n_out=1),
      tl.Dropout(rate=dropout, mode=mode)
      )


def _ror(x, n, p=1):
  """Bitwise right rotation.

  Args:
    x: np.array
    n: Bit count to represent each value of x
    p: Bit positions to shift

  Returns:
    np.array: x with all values shifted by p positions in n bits
  """
  a = np.right_shift(x, p)
  b = np.left_shift(1, p) - 1
  c = np.bitwise_and(x, b)
  d = np.left_shift(c, n - p)

  return a + d


def _rol(x, n, p=1):
  """Bitwise left rotation.

  Args:
    x: np.array
    n: Bit count to represent each value of x
    p: Bit positions to shift

  Returns:
    np.array: x with all values shifted by p positions in n bits
  """
  a = np.left_shift(x, p)
  b = np.left_shift(1, n) - 1
  c = np.bitwise_and(a, b)
  d = np.right_shift(x, n - p)

  return np.bitwise_or(c, d)


def _shuffle_layer(inputs, shuffle_fn):
  """Shuffles the elements according to bitwise left or right rotation.

  Args:
    inputs: Tensor input from previous layer
    shuffle_fn: Shift function rol or ror

  Returns:
    tf.Tensor: Inputs shifted according to shuffle_fn
  """
  seq_length = inputs.shape[1]
  n_bits = np.int32(np.log(seq_length - 1) / np.log(2.0)) + 1

  indices = np.arange(0, seq_length).astype('int32')
  rev_indices = shuffle_fn(indices, n_bits)
  return jnp.take(inputs, rev_indices, axis=1, mode='clip')


@assert_shape('bld->bld')
def ShuffleLayer():
  return tl.Fn(
      'ShuffleLayer', lambda x: _shuffle_layer(x, _rol), n_out=1)


@assert_shape('bld->bld')
def ReverseShuffleLayer():
  return tl.Fn(
      'ReverseShuffleLayer', lambda x: _shuffle_layer(x, _ror), n_out=1)


@assert_shape('...,bld->...,bld')
def _ForwardStep(d_model, dropout, mode):
  """Takes (n_layer, state) and returns (n_layer, shuffle_layer(rsu(state)))."""
  return tl.Parallel([], tl.Serial(
      ResidualSwitchUnit(d_model, dropout, mode),
      ShuffleLayer(),
  ))


@assert_shape('...,bld->...,bld')
def _BackwardStep(d_model, dropout, mode):
  """Takes (n_layer, state) and returns (n_layer, reverse_shuffle_layer(rsu(state)))."""
  return tl.Parallel([], tl.Serial(
      ResidualSwitchUnit(d_model, dropout, mode),
      ReverseShuffleLayer(),
  ))


@assert_shape('bld->bld')
def BenesBlock(d_model, dropout, mode):
  def bit_sequence(inputs):
    seq_length = inputs.shape[1]
    n_bits = np.int32(np.log(seq_length - 1) / np.log(2.0)) + 1
    return jnp.arange(0, n_bits)
  return tl.Serial(
      tl.Dup(),
      tl.Fn('BitSeq', bit_sequence, n_out=1),
      tl.Scan(_ForwardStep(d_model, dropout, mode)),
      tl.Scan(_BackwardStep(d_model, dropout, mode)),
      tl.Select([1]),
  )


@assert_shape('bl->blv')
def ResidualShuffleExchange(vocab_size,
                            d_model,
                            input_dropout,
                            dropout,
                            mode='train',
                            n_blocks=2):
  """Returns a Residual Shuffle Exchange Network model."""
  benes_blocks = [BenesBlock(d_model, dropout, mode) for _ in range(n_blocks)]
  return tl.Serial(
      tl.Embedding(vocab_size, d_model),
      tl.Dropout(rate=input_dropout, mode=mode),
      # Apply Benes Block n_blocks times.
      *benes_blocks,
      ResidualSwitchUnit(d_model, dropout, mode),
      # Produce probabilities.
      tl.Dense(vocab_size),
      tl.LogSoftmax(),
  )
